/*
 *   Msp3410D soundcard
 *   Copyright (c) by Wilfried Philips <philips@telin.UGent.be>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc.,  59 Temple Place, Suite 330 
 *   Boston, MA  02111-1307 USA 
 *
 */

#define SND_MAIN_OBJECT_FILE
#define __SND_OSS_COMPAT__
#include "include/driver.h"
#include "include/pcm.h"
#include "include/mixer.h"
#include "include/rawmidi.h"
#include "include/initval.h"
#include "msputil.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) 
/*#include "wphkcompat24.h"*/
#endif
#include "linux/videodev.h"

static int debug = 0;




 
typedef int (*msp_command_proc)(struct i2c_client *client, unsigned int cmd, void *arg);


typedef struct snd_card_msp3410D {
  struct i2c_client*i2c_client;
  char id[64];
  snd_card_t *card;
  msp_command_proc orig_msp_command;
  msp_shadow shadow;
  snd_kmixer_t *mixer;
  snd_kmixer_element_t *io_loudspeaker;
  snd_kmixer_element_t *sw1_volume_onoff_switch;
  int volume_off;
  snd_kmixer_element_t *vol_main_volume;
  int main_volume_called;
  /*  snd_kmixer_element_t *accu_main_out;*/
  snd_kmixer_element_t *sw1_loudness_onoff_switch;
  int loudness_off;
  snd_kmixer_element_t *sw1_loudness_super_switch;
  snd_kmixer_element_t *loudness_strength;
  snd_kmixer_element_t *sw1_effect3D_onoff_switch;
  int effect3D_off;
  snd_kmixer_element_t *sw1_effect3D_mode_switch;
  snd_kmixer_element_t *effect3D_strength;
  snd_kmixer_element_t *effect3D_highpass_gain;
  snd_kmixer_element_t *sw1_effect3D_highpass_auto;
  short highpass_auto_off;
  /*  snd_kmixer_element_t *accu_central_in;*/
  snd_kmixer_element_t *sw1_beeper_onoff_switch;
  int beeper_off;
  snd_kmixer_element_t* beeper_volume;
  snd_kmixer_element_t* beeper_freq;
  snd_kmixer_element_t *sw1_equalizer_switch;
#ifdef EQ_ONE_GROUP
  snd_kmixer_element_t *tone_main_equalizer_tone_control;
  snd_kmixer_element_t *tone_main_basstreble_tone_control;
#else
  snd_kmixer_element_t *tone_main_equalizer_tone_control[5];
  snd_kmixer_element_t *tone_main_basstreble_tone_control[2];
#endif
  snd_kmixer_element_t *io_video_in;
  snd_kmixer_element_t *sw1_avc_switch;
  snd_kmixer_element_t *avc_decay_time;
  spinlock_t mixer_lock;
  /*checked above here*/

} snd_card_msp3410D_t;



#define MAXMSPS SND_CARDS


/* currently registered cards */
static  int nmsps=0;		
static snd_card_msp3410D_t *mspcards[MAXMSPS];



#define SND_CARD_TYPE_MSP3410D SND_CARD_TYPE_DUMMY


/*overrides the index of the sound card as used in 
  modules.conf: snd-card-7*/

#if 0				/* wph: turn off for now */
int snd_index[SND_CARDS] = SND_DEFAULT_IDX;	/* Index 0-MAX */

/*overrides the textual id for the card*/
char *snd_id[SND_CARDS] = SND_DEFAULT_STR;	/* ID for this card */

/*override: switches of one of the cards*/
int snd_enable[SND_CARDS] = {1, [1 ... (SND_CARDS - 1)] = 0};

#endif


int msp_register_unregister_msp(struct i2c_client *c);
int alsa_msp3410D_module_init(void);
int delete_mspcard(snd_card_msp3410D_t *msp);
snd_card_msp3410D_t *create_mspcard(int i, struct i2c_client *c);
void alsa_msp3410D_module_exit(void);
int wph_test(struct i2c_client *client, int *voices);
int delete_mixer(snd_card_msp3410D_t *msp);
int create_mixer(snd_card_msp3410D_t *msp3410D, int device);


/* ----------------------------------------------------------------------- */

static int msp_adapters(struct i2c_adapter *adap);
static int msp_clients(struct i2c_client *client);

#ifndef I2C_DRIVERID_CARD_MSP3410D
#define I2C_DRIVERID_CARD_MSP3410D I2C_DRIVERID_EXP2
#endif

static struct i2c_driver driver = {
	name: "msp3410d mixer",
        id:   I2C_DRIVERID_CARD_MSP3410D /* I2C_DRIVERID_FIXME */,
	flags: I2C_DF_DUMMY,
        attach_adapter: msp_adapters,
        detach_client: msp_clients,
};


static void snd_card_msp3410D_use_inc(snd_card_t * card)
{
  int i;
  struct i2c_client*client;
  MOD_INC_USE_COUNT;
  for(i=0;i<MAXMSPS;i++)
    {
      if(mspcards[i]->card==card)
	break;
    }
  if(i==MAXMSPS)
    {
      eprk(__FUNCTION__ ": card not found\n");
      return;
    }
  client=mspcards[i]->i2c_client;
  if(client==0)
    {
      eprk(__FUNCTION__ ":client=0\n");
      return;
    }
  /*the following line locks the driver for this card in memory*/
  if (client->adapter->inc_use)
    client->adapter->inc_use(client->adapter);

}

static void snd_card_msp3410D_use_dec(snd_card_t * card)
{
  int i;
  struct i2c_client*client;
  MOD_DEC_USE_COUNT;
  for(i=0;i<MAXMSPS;i++)
    {
      if(mspcards[i]->card==card)
	break;
    }
  if(i==MAXMSPS)
    {
      eprk(__FUNCTION__ ": card not found\n");
      return;
    }
  client=mspcards[i]->i2c_client;
  if(client==0)
    {
      eprk(__FUNCTION__ ":client=0\n");
      return;
    }
  /*the following line unlocks the driver for this card in memory*/
  if (client->adapter->dec_use)
    client->adapter->dec_use(client->adapter);
}





static int msp_adapters(struct i2c_adapter *adap)
{
  int i;
  /*wphid;*/
 if (debug)
    printk("card-msp3410d: adapter %s\n",adap->name);
  for (i=0; i<I2C_CLIENT_MAX; i++) {
    if (!adap->clients[i])
      continue;
    msp_clients(adap->clients[i]);
  }
  return 0;
}





static int intercept_msp_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
  int i;
  snd_card_msp3410D_t *msp=0;
  dprk(__FUNCTION__ ": called with client=0x%x cmd=%d\n",(int)client,cmd);
  for(i=0;i<MAXMSPS;i++)
    {
      if(mspcards[i]->i2c_client==client)
	break;
    }
  if(i==MAXMSPS)
    {
      eprk(__FUNCTION__ ": could not locate msp of this client\n");
      return -1;
    }
  msp=mspcards[i];
  switch(cmd)
    {
    case VIDIOCGAUDIO:
      wphid;/**/
      break;
    case VIDIOCSAUDIO:
      wphid;/**/
      {      

	/* programs like xawtv tend to ignore mixers and
	   set audio volumes themselves.
	   we substitue fake values here; this effectively means
	   that volumes can only be changed through our driver
	 */
#if 0
	struct video_audio *va = arg;
	int low=(unsigned char)0x0;
	int high=(unsigned char)0x73;
	int vol,balance,bal;
	wph_mspget_shadow(client,0x0000,WPH_DSP,&vol,0xfff0,&msp->shadow);
	va->volume=((vol-low)*65536)/(high-low);
	wph_mspget_shadow(client,0x0001,WPH_DSP,&balance,HIGH,&msp->shadow);
	bal=balance=(signed char) balance;
	balance=(balance>=0)?127-balance:balance+127;
	balance=(balance*32768)/127;
	balance=(bal>0)?65535-balance:balance;
#endif
	return 0;
      }
      break;
    case VIDIOCSCHAN:
      wphid;
      break;
    case VIDIOCSFREQ:
      wphid;/**/
      break;      
    }
  if(!msp->orig_msp_command)
    {
      eprk("msp->orig_msp_command=0\n");
      return -1;
    }
  dprk(__FUNCTION__ ": end of call with client=0x%x cmd=%d\n",(int)client,cmd);
  return (msp->orig_msp_command)(client,cmd,arg);
}



int   msp_register_unregister_msp(struct i2c_client *c)
{
  int i;
  /*if client already exists it should be unregistered*/
  for (i = 0; i < MAXMSPS; i++) 
    {
      if (!mspcards[i]
	  ||mspcards[i]->i2c_client !=c) continue;
#if 1
      /*we stop intercepting commands to the msp driver 
	because we want to know when the msp3400c 
	driver from bttv changes settings such as 
	volume levels
      */

      c->driver->command= mspcards[i]->orig_msp_command; 
      mspcards[i]->orig_msp_command=0; 
#endif
      dprk("client %s unregistered\n",c->name);
      delete_mspcard(mspcards[i]);
      mspcards[i]=0;
      nmsps--;
      if(nmsps<0)
	{
	  eprk("unregistered more than registered\n");
	  nmsps=0;
	  return -1;
	}
      return 0;
    }
  /*if client does not exist it should be registered*/
  for(i=0;i<MAXMSPS;i++)
    {
      if (mspcards[i]) continue;
      mspcards[i]=create_mspcard(i,c);      
      if(!mspcards[i])
	{
	  eprk("could not register %s\n",c->name);
	  return -1;
	}
#if 1
      /*we intercept commands to the msp driver 
	because we want to know when the msp3400c 
	driver from bttv changes settings such as 
	volume levels
      */
      mspcards[i]->orig_msp_command=c->driver->command; 
      c->driver->command= &intercept_msp_command;
#endif
      nmsps++;
      dprk("wph: %s (%s) registered\n",
	   c->name,c->adapter->name); 
      return 0;
    }
  eprk("too many msps\n");
  return -1;
}

static int msp_clients(struct i2c_client *client)
{
  /* TV card ??? */
  if (client->adapter->id != (I2C_ALGO_BIT | I2C_HW_B_BT848)) {
    dprk(" %s is not a tv card\n",
	 client->adapter->name);
    return -1;
  }
  dprk("msp3410d: examining  %s\n",client->name);
  if(client->driver->id!=I2C_DRIVERID_MSP3400)
    {
      dprk("card-msp3410d: %s is not an MSP i\n",
	   client->driver->name);
      return -1;
      
    }
  msp_register_unregister_msp(client);
  return 0;
}

int alsa_msp3410D_module_init(void)
{
  /*wphid;*/
  memset(&mspcards,0,sizeof(mspcards));
  nmsps=0;
  i2c_add_driver(&driver);
  return 0;
}


static void snd_card_msp_free(snd_card_msp3410D_t *msp)
{
  if (msp == NULL)
    return;
  snd_card_unregister(msp->card);
  snd_magic_kfree(msp);
}



int delete_mspcard(snd_card_msp3410D_t*msp)
{
  dprk("delete_mspcard %s\n",msp->id);
#if 0				/* wph: not required */
  if(delete_mixer(msp))
    {
      eprk("failed to delete mixer\n");
      return -1;
    }
#endif
  snd_card_msp_free(msp);
#if 0				/* wph1 */
  snd_card_free(msp->card);
  snd_magic_kfree(msp); 
#endif

  return 0;
}

snd_card_msp3410D_t* create_mspcard(int i, struct i2c_client*c)
{
  int err=0;
  snd_card_t *card;
  snd_card_msp3410D_t*msp;
  dprk("create_mspcard: i=%d\n",i);
  msp = snd_magic_kcalloc(snd_card_msp3410D_t, 0, GFP_KERNEL);
  if (msp == NULL)    {
      eprk("unable to allocate msp3410D structure\n");
      return 0;
  }
  msp->i2c_client=c;
  sprintf(msp->id,"%s%d",c->name,c->adapter->name);
  card = snd_card_new(-1,0,snd_card_msp3410D_use_inc, 
		      snd_card_msp3410D_use_dec);
  if (card == NULL) {
    snd_card_msp_free(msp);
    msp=0;
    eprk("snd_card_new failed for %s\n",c->name);
    return 0;
  }
  msp->card = card;
  card->type = SND_CARD_TYPE_MSP3410D;

  if (create_mixer(msp, 0))
    {
      snd_card_msp_free(msp);
      msp=0;
      eprk("create_mixer failed for %s\n",c->name);
      return 0;      
    }
  strcpy(card->abbreviation, "MSP3410D");
  strcpy(card->shortname, "MSP3410D");
  sprintf(card->longname, "MSP3410D %i", i);
  if((err=snd_card_register(card)))
    {
      eprk("unable to register card %s: %d\n",card->shortname, 
	   err);
      return 0;
    }
  dprk("probe end\n");
  return msp;
}


void alsa_msp3410D_module_exit(void)
{
  int i;
  /*wphid;*/
  i2c_del_driver(&driver);
  for(i=0;i<MAXMSPS;i++)
    {
      if(!mspcards[i]) continue;
      delete_mspcard(mspcards[i]);      
      mspcards[i]=0;
      nmsps--;      
    }
  if(nmsps)
    eprk("nmsps=%d and should be 0\n",nmsps);
}

/*---------------------------------------------------*/








snd_card_msp3410D_t * msp_of_element(snd_kmixer_element_t *element)
{
  snd_card_msp3410D_t *msp=0;
  msp=element->private_data;
  if(msp==0)
    {
      eprk("element.private_data does not conatin msp: element=%x ",
	   (int)element);
      return 0;
    }
  else 
    {
      /*dprk("located msp (%s) of element\n",msp->id);*/
    }
  
  return msp;
}


int cb_volume_onoff_switch(snd_kmixer_element_t *element, 
			int w_flag, unsigned int *bitmap)
{
  int changed=0;
  unsigned long flags;
  snd_card_msp3410D_t *msp=0;
  /*dprk(__FUNCTION__ ": w=%d value=%d\n",bitmap[0],w_flag);*/
  msp=msp_of_element(element);
  if(msp==0) return 0;
  /*dprk("located switch on card %s\n",msp->id);*/
  if (element!=msp->sw1_volume_onoff_switch)
    {
      /*dprk(__FUNCTION__ ": element %x not recognised\n", 
	(int)element);	*/
      /* this sometimes happens
				   but should not cause an error*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=snd_mixer_get_bit(bitmap,0);
      if(!val)
	{
	  /* turn effect off but do not write shadow register */
	  wph_msplockandset(msp->i2c_client,0x0000,WPH_DSP,0xffe,0xfff0,
			    &msp->shadow);
	}
      else
	{
	  wph_mspunlock(msp->i2c_client,0x0000,WPH_DSP,0xfff0,
			    &msp->shadow); 
	  /*dprk("HERE: val=%d\n",val);*/
	}
      changed=msp->volume_off==val; /* == is no mistake */
      msp->volume_off=!val;
    }
  else
    {
      snd_mixer_set_bit(bitmap,0,!msp->volume_off);
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}

static int cb_main_volume(snd_kmixer_element_t * element, 
			  int w_flag, int *voices)
{
  int changed=0;
  int chiplow=(unsigned char)0x0; /*  (minumum for chip) */
  int low= (unsigned  char) 0x37;	 /* -60dB (our minimum) */
  int high= (unsigned  char) 0x73; /* 0dB */
  snd_card_msp3410D_t*msp=0;
  unsigned long flags;
  int clipmode=0;
  int mute=0;
  /*  wphid;
  dprk("main_volume w_flag=%d voices=%d %d\n",w_flag,voices[0],voices[1]);
  */
  msp=msp_of_element(element);
  if(!msp)return -1;
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int left=voices[0];
      int right=voices[1];
      int volume=0;
      int balance=0;
      if(!msp->main_volume_called)
	{
	  msp->main_volume_called=1;
	  /* alsa mutes all cards when module is loaded
	   we wish to ignore this*/
	  goto __end;
	}
      if(left>right)
	balance=(right*127)/left;
      else if(right!=0) 
	  balance=left*127/right;
      if(balance<0||balance>127)
	{
	  eprk("range error: balance=%d\n",balance);
	  if(balance>127)balance=127; else balance=0;
	}
      balance=(left<right)?127-balance:balance-127;
      volume=(left<right)?right:left;
      volume=((high-low)*volume)/100+low;
      if(volume<low||volume>high)
	{
	  eprk("range error: volume=%d\n",volume);
	  if(volume>high)volume=high; else volume=low;
	}
      /* set balance value and linear balance mode */
      /*wphid;*/
      if(volume==low) volume=chiplow;		/* we do a real mute */
      changed=
	wph_mspset(msp->i2c_client,0x0001,WPH_DSP,
		   balance<<8,HIGH|LOW,&msp->shadow);
      changed |=wph_mspset(msp->i2c_client,0x0000,WPH_DSP,
			   (volume<<8)|clipmode,HIGH|LOW,&msp->shadow);      
    }
  else				/* w_flag=0 */
    {
      int balance=0,volume=0,left=0,right=0;
      wph_mspget_shadow(msp->i2c_client,0x0000,WPH_DSP,&volume,0xffff,
			&msp->shadow); 
      clipmode=(volume&0x3);
      volume>>=8;
      if(volume==0xffe)
	{ mute=1;volume=low;}
      if(volume<low)volume=low;
      /*dprk("volume is now %x\n",volume);*/
      volume=((volume-low)*100)/(high-low);
      if(volume<0||volume>100)
	{
	  eprk("range error: volume=%d\n",volume);
	  if(volume>100)volume=100; else volume=0;
	}
      /* we hope that balance is in linear mode */
      wph_mspget_shadow(msp->i2c_client,0x0001,WPH_DSP,&balance,HIGH,&msp->shadow);
      balance=(signed char)balance;
      if(balance<0)
	{
	  right=-1;		/* save sign of balance */
	}
      balance=(balance>=0)?127-balance:balance+127;
      if(right==-1)
	{
	  left=volume;
	  right=(left*balance)/127;
	  if(right<0||right>100)
	    {
	      eprk("range error: right=%d\n",right);
	      if(right>100)right=100; else right=0;
	    }
	}
      else
	{
	  right=volume;
	  left=(right*balance)/127;
	  if(left<0||left>100)
	    {
	      eprk("range error: left=%d\n",left);
	      if(left>100)left=100; else left=0;
	    }
	}
      voices[0]=left;voices[1]=right;
    }
 __end:
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}



int vol_control (snd_card_msp3410D_t *msp,
		 int address,int mask,
		 int dev_low, int dev_high, 
		 int drv_low, int drv_high,int w_flag,
		 int *dev_val,int *drv_val)
{
  int changed=0;
  unsigned long flags;
  if(!msp) return -1;
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val= *drv_val;	/* between drv_low and drv_high */
      val=((dev_high-dev_low)*(val-drv_low))
	/(drv_high-drv_low)+dev_low;
      if(val<dev_low||val>dev_high)
	{
	  eprk("range error: val=%d low=%d high=%d\n",val,dev_low,dev_high);
	  if(val>dev_high)val=dev_high; else val=dev_low;
	}
      *dev_val=val;
      changed=wph_mspset(msp->i2c_client,address,
			 WPH_DSP,val,mask,&msp->shadow);
    }
  else
    {
      int val=0;
      wph_mspget_shadow(msp->i2c_client,address,WPH_DSP,&val,
			mask,&msp->shadow);
      if(dev_low<0)
	val=(signed char)val;
      else
	val=(unsigned char)val;
      *dev_val=val;
      val=((drv_high-drv_low)*(val-dev_low))/
      (dev_high-dev_low)+drv_low;
      if(val<drv_low||val>drv_high)
	{
	  eprk("range error: val=%d low=%d high=%d\n",val,drv_low,drv_high);
	  if(val>drv_high)val=drv_high; else val=drv_low;
	}
      *drv_val=val;
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}


#ifdef EQ_ONE_GROUP

int cb_main_equalizer_tone_control (snd_kmixer_element_t *element,int w_flag, 
				    int *voices)
{
  int i;
  int n=0;
  int address=0;
  int changed=0;
  int dev_val=0;
  snd_card_msp3410D_t *msp=0;
  int low= (signed char) 0xa0;			/* range of amplification */
  int high= (signed char) 0x60;
  /* dprk("equalizer_tone_control: elememt=0x%x write=%d voices[0]=%d\n",
    (int)element, w_flag,voices[0]);
  */
  msp=msp_of_element(element);
  if(!msp) return -1;
  if(element==msp->tone_main_equalizer_tone_control)
    {
      /*dprk("element is number %d\n",i);*/
      address=0x21;
      n=5;
    }
  else
    {
      /*eprk("element not found\n");*/
      /* this sometimes happend, but I don't know if it
	 is allowed to happen*/
      return 0;
    }
  for(i=0;i<n;i++)
    changed|=vol_control (msp,address+i,HIGH,
			  low, high,
			  -50,50,w_flag,&dev_val,&voices[i]);
  return changed;
}

int cb_main_basstreble_tone_control (snd_kmixer_element_t *element,int w_flag, 
				    struct snd_mixer_element_tone_control1*tc1)
{

  int n=0;
  int address=0;
  int changed=0;
  int dev_val=0;
  snd_card_msp3410D_t *msp=0;
  int low= (signed char) 0xa0;			/* range of amplification */
  int high= (signed char) 0x60;
  /* dprk("equalizer_tone_control: elememt=0x%x write=%d voices[0]=%d\n",
    (int)element, w_flag,voices[0]);
  */
  msp=msp_of_element(element);
  if(!msp) return -1;
    if(element==msp->tone_main_basstreble_tone_control)
      {
	/*dprk("element is number %d\n",i);*/
	address=0x02;
	n=2;
      }
  
    else
    {
      /*eprk("element not found\n");*/
      /* this sometimes happend, but I don't know if it
	 is allowed to happen*/
      return 0;
    }
    if(tc1->tc&SND_MIXER_TC1_BASS)
      changed|=vol_control (msp,address,HIGH,
			    low, high,
			    -50,50,w_flag,&dev_val,&tc1->bass);
    if(tc1->tc&SND_MIXER_TC1_TREBLE)
      changed|=vol_control (msp,address+1,HIGH,
			    low, high,
			    -50,50,w_flag,&dev_val,&tc1->treble);
    return changed;
}
#else
int cb_main_equalizer_tone_control (snd_kmixer_element_t *element,int w_flag, 
				    int *voices)
{
  int i;
  int number=-1;
  int address=0;
  int changed=0;
  int is_equalizer=0;
  int is_basstreble=0;
  snd_card_msp3410D_t *msp=0;
  int low= (signed char) 0xa0;			/* range of amplification */
  int high= (signed char) 0x60;
  unsigned long flags;
  /* dprk("equalizer_tone_control: elememt=0x%x write=%d voices[0]=%d\n",
    (int)element, w_flag,voices[0]);
  */
  msp=msp_of_element(element);
  if(!msp) return -1;
  for(i=0;i<5;i++) 
    if(element==msp->tone_main_equalizer_tone_control[i])
      {
	/*dprk("element is number %d\n",i);*/
	number=i;
	is_equalizer=1;
	address=0x21+number;
      }
  if(!is_equalizer)
    for(i=0;i<2;i++) 
      if(element==msp->tone_main_basstreble_tone_control[i])
	{
	  /*dprk("element is number %d\n",i);*/
	  number=i;
	  is_basstreble=1;
	  address=0x02+number;
	}

  if(!is_equalizer&&!is_basstreble)
    {
      /*eprk("element not found\n");*/
      /* this sometimes happend, but I don't know if it
	 is allowed to happen*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=voices[0];	/* between -50 and 50 */
      val=((high-low)*(val+50))/100+low;
      if(val<low||val>high)
	{
	  eprk("range error: val=%d\n",val);
	  if(val>high)val=high; else val=low;
	}
      changed=wph_mspset(msp->i2c_client,address,
			 WPH_DSP,val,HIGH,&msp->shadow);
    }
  else
    {
      int val=0;
      wph_mspget_shadow(msp->i2c_client,address,WPH_DSP,&val,HIGH,&msp->shadow);
      val=(signed char)val;
      val=(100*(val-low))/(high-low)-50;
      if(val<-50||val>50)
	{
	  eprk("range error: val=%d\n",val);
	  if(val>50)val=50; else val=-50;
	}
      voices[0]=val;
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}
#endif


int cb_beeper_onoff_switch(snd_kmixer_element_t *element, 
			int w_flag, unsigned int *bitmap)
{
  int changed=0;
  unsigned long flags;
  snd_card_msp3410D_t *msp=0;
  /*dprk(__FUNCTION__ ": w=%d value=%d\n",bitmap[0],w_flag);*/
  msp=msp_of_element(element);
  if(msp==0) return 0;
  /*dprk("located switch on card %s\n",msp->id);*/
  if (element!=msp->sw1_beeper_onoff_switch)
    {
      /*dprk(__FUNCTION__ ": element %x not recognised\n", 
	(int)element);	*/
      /* this sometimes happens
				   but should not cause an error*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=snd_mixer_get_bit(bitmap,0);
      /* turn effect off, but do not write shadow register */
      if(!val)
	{
	  wph_msplockandset(msp->i2c_client,0x0014,WPH_DSP,0x00,HIGH,
			    &msp->shadow);
	}
      else
	{
	  wph_mspunlock(msp->i2c_client,0x0014,WPH_DSP,HIGH,
			&msp->shadow); 
	  /*dprk("HERE: ret=%d\n",ret);*/
	}
      changed=msp->beeper_off==val; /* == is no mistake */
      msp->beeper_off=(!val);
    }
  else
    {
      snd_mixer_set_bit(bitmap,0,!msp->beeper_off);
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}

int cb_beeper_volume (snd_kmixer_element_t *element,int w_flag, 
				    int *voices)
{
  int changed=0;
  snd_card_msp3410D_t *msp=0;
  int low= (unsigned char) 0x00; /* effect off */
  int high= (unsigned char) 0x7f;	/* max volume */
  unsigned long flags;
  /*dprk(__FUNCTION__ ": element=0x%x write=%d voices[0]=%d\n",
    (int)element, w_flag,voices[0]);*/
  msp=msp_of_element(element);
  if(!msp) return -1;
  if(element!=msp->beeper_volume)
    {
      /*eprk("element not found\n");*/
      /* this sometimes happens, but I don't know if it
	 is allowed to happen*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=voices[0];	/* between 0 and 100 */
      val=((high-low)*val)/100+low;
      if(val<low||val>high)
	{
	  eprk("range error: val=%d\n",val);
	  if(val>high)val=high; else val=low;
	}
      changed=wph_mspset(msp->i2c_client,0x0014,WPH_DSP,val,HIGH,&msp->shadow);
    }
  else
    {
      int val=0;
      wph_mspget_shadow(msp->i2c_client,0x0014,WPH_DSP,&val,HIGH,&msp->shadow);
      val=(signed char)val;
      val=(100*(val-low))/(high-low);
      if(val<0||val>100)
	{
	  eprk("range error: val=%d\n",val);
	  if(val>100)val=100; else val=0;
	}
      voices[0]=val;
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}


int cb_beeper_freq (snd_kmixer_element_t *element,int w_flag, 
				    int *voices)
{
  int changed=0;
  snd_card_msp3410D_t *msp=0;
  int low= (unsigned char) 0x01; /* 16 Hz */
  int high= (unsigned char) 0xff;	/* 4kHz */
  unsigned long flags;
  /*dprk(__FUNCTION__ ": element=0x%x write=%d voices[0]=%d\n",
    (int)element, w_flag,voices[0]);*/
  msp=msp_of_element(element);
  if(!msp) return -1;
  if(element!=msp->beeper_freq)
    {
      /*eprk("element not found\n");*/
      /* this sometimes happend, but I don't know if it
	 is allowed to happen*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=voices[0];	/* between 0 and 100 */
      val=((high-low)*val)/100+low;
      if(val<low||val>high)
	{
	  eprk("range error: val=%d\n",val);
	  if(val>high)val=high; else val=low;
	}
      changed=wph_mspset(msp->i2c_client,0x0014,WPH_DSP,val,LOW,&msp->shadow);
    }
  else
    {
      int val=0;
      wph_mspget_shadow(msp->i2c_client,0x0014,WPH_DSP,&val,LOW,&msp->shadow);
      val=(signed char)val;
      val=(100*(val-low))/(high-low);
      if(val<0||val>100)
	{
	  eprk("range error: val=%d\n",val);
	  if(val>100)val=100; else val=0;
	}
      voices[0]=val;
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}

int cb_loudness_strength (snd_kmixer_element_t *element,int w_flag, 
				    int *voices)
{
  int changed=0;
  snd_card_msp3410D_t *msp=0;
  int low= (signed char) 0x00; /* effect off */
  int high= (signed char) 0x44;	/* 17dB boost */
  unsigned long flags;
  /*dprk("loudness_strength: element=0x%x write=%d voices[0]=%d\n",
    (int)element, w_flag,voices[0]);*/
  msp=msp_of_element(element);
  if(!msp) return -1;
  if(element!=msp->loudness_strength)
    {
      /*eprk("element not found\n");*/
      /* this sometimes happend, but I don't know if it
	 is allowed to happen*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=voices[0];	/* between 0 and 100 */
      val=((high-low)*val)/100+low;
      if(val<low||val>high)
	{
	  eprk("range error: val=%d\n",val);
	  if(val>high)val=high; else val=low;
	}
      changed=wph_mspset(msp->i2c_client,0x0004,WPH_DSP,
			 val,HIGH,&msp->shadow);
    }
  else
    {
      int val=0;
      wph_mspget_shadow(msp->i2c_client,0x0004,WPH_DSP,&val,HIGH,&msp->shadow);
      val=(signed char)val;
      val=(100*(val-low))/(high-low);
      if(val<0||val>100)
	{
	  eprk("range error: val=%d\n",val);
	  if(val>100)val=100; else val=0;
	}
      voices[0]=val;
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}




int cb_loudness_onoff_switch(snd_kmixer_element_t *element, 
			int w_flag, unsigned int *bitmap)
{
  int changed=0;
  unsigned long flags;
  snd_card_msp3410D_t *msp=0;
  /*dprk("loudness_super_switch: w=%d value=%d\n",bitmap[0],w_flag);*/
  msp=msp_of_element(element);
  if(msp==0) return 0;
  /*dprk("located switch on card %s\n",msp->id);*/
  if (element!=msp->sw1_loudness_onoff_switch)
    {
      /*dprk(__FUNCTION__ ": element %x not recognised\n", 
	(int)element);	*/
      /* this sometimes happens
				   but should not cause an error*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=snd_mixer_get_bit(bitmap,0);
      /* turn effect off, but do not write shadow register */
      if(!val)
	wph_msplockandset(msp->i2c_client,0x0004,WPH_DSP,0x00,HIGH,
			  &msp->shadow);
      else
	{
	  /*wat als effect door mixer wordt aangezet, terwijl er nooit 
	    voordien naar register werd geschreven?*/
	  wph_mspunlock(msp->i2c_client,0x0004,WPH_DSP,HIGH,
			    &msp->shadow); 
	  /*dprk("HERE: val=%d\n",val);*/
	}
      changed=msp->loudness_off==val; /* == is no mistake */
      msp->loudness_off=(!val);
    }
  else
    {
      snd_mixer_set_bit(bitmap,0,!msp->loudness_off);
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}

int cb_loudness_super_switch(snd_kmixer_element_t *element, 
			int w_flag, unsigned int *bitmap)
{
  int changed=0;
  unsigned long flags;
  snd_card_msp3410D_t *msp=0;
  /*dprk("loudness_super_switch: w=%d value=%d\n",bitmap[0],w_flag);*/
  msp=msp_of_element(element);
  if(msp==0) return 0;
/*dprk("located switch on card %s\n",msp->id);*/
  if (element!=msp->sw1_loudness_super_switch)
    {
      /*dprk(__FUNCTION__ ": element %x not recognised\n", 
	(int)element);	*/
      /* this sometimes happens
				   but should not cause an error*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=bitmap[0]?0x4:0x00;
      changed=wph_mspset(msp->i2c_client,0x04,WPH_DSP,val,LOW,&msp->shadow);
      
    }
  else
    {
      int val;
      wph_mspget_shadow(msp->i2c_client,0x04,WPH_DSP,&val,LOW,&msp->shadow);
      bitmap[0]=(val==0x4);
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}


int cb_effect3D_onoff_switch(snd_kmixer_element_t *element, 
			int w_flag, unsigned int *bitmap)
{
  int changed=0;
  unsigned long flags;
  snd_card_msp3410D_t *msp=0;
  /*dprk(__FUNCTION__ ": w=%d value=%d\n",bitmap[0],w_flag);*/
  msp=msp_of_element(element);
  if(msp==0) return 0;
  /*dprk("located switch on card %s\n",msp->id);*/
  if (element!=msp->sw1_effect3D_onoff_switch)
    {
      /*dprk(__FUNCTION__ ": element %x not recognised\n", 
	(int)element);	*/
      /* this sometimes happens
				   but should not cause an error*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=snd_mixer_get_bit(bitmap,0);
      /* turn effect off, but do not write shadow register */
      if(!val)
	wph_msplockandset(msp->i2c_client,0x0005,WPH_DSP,0x00,HIGH,
			  &msp->shadow);
      else
	{
	  /*wat als effect door mixer wordt aangezet, terwijl er nooit 
	    voordien naar register werd geschreven?*/
	  wph_mspunlock(msp->i2c_client,0x0005,WPH_DSP,HIGH,
			    &msp->shadow); 
	  /*dprk("HERE: ret=%d\n",ret);*/
	}
      changed=msp->effect3D_off==val; /* == is no mistake */
      msp->effect3D_off=(!val);
    }
  else
    {
      snd_mixer_set_bit(bitmap,0,!msp->effect3D_off);
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}

int cb_effect3D_strength (snd_kmixer_element_t *element,int w_flag, 
				    int *voices)
{
  int changed=0;
  snd_card_msp3410D_t *msp=0;
  int low= (signed char) 0x80; /* almost 100% reduction */
  int high= (signed char) 0x7f;	/* 100% enhancement */
  unsigned long flags;
  /*dprk("effect3D_strength: element=0x%x write=%d voices[0]=%d\n",
    (int)element, w_flag,voices[0]);*/
  msp=msp_of_element(element);
  if(!msp) return -1;
  if(element!=msp->effect3D_strength)
    {
      /*eprk("element not found\n");*/
      /* this sometimes happend, but I don't know if it
	 is allowed to happen*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=voices[0];	/* between -50 and 50 */
      val=((high-low)*(val+50))/100+low;
      if(val<low||val>high)
	{
	  eprk("range error: val=%d\n",val);
	  if(val>high)val=high; else val=low;
	}
      changed=wph_mspset(msp->i2c_client,0x0005,WPH_DSP,val,HIGH,&msp->shadow);
    }
  else
    {
      int val=0;
      wph_mspget_shadow(msp->i2c_client,0x0005,WPH_DSP,&val,HIGH,&msp->shadow);
      val=(signed char)val;
      val=(100*(val-low))/(high-low)-50;
      if(val<-50||val>50)
	{
	  eprk("range error: val=%d\n",val);
	  if(val>50)val=50; else val=-50;
	}
      voices[0]=val;
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}

int cb_effect3D_highpass_gain (snd_kmixer_element_t *element,int w_flag, 
				    int *voices)
{
  int changed=0;
  int low=0;
  int high=6;
  int autogain=8;
  snd_card_msp3410D_t *msp=0;
  unsigned long flags;
  /*dprk(__FUNCTION__ ": element=0x%x write=%d voices[0]=%d\n",
    (int)element, w_flag,voices[0]);*/
  msp=msp_of_element(element);
  if(!msp) return -1;
  if(element!=msp->effect3D_highpass_gain)
    {
      /*eprk("element not found\n");*/
      /* this sometimes happend, but I don't know if it
	 is allowed to happen*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=high-voices[0];	/* between 0 and 6 */
      if(val<low||val>high)
	{
	  eprk("range error: val=%d\n",val);
	  if(val>high)val=high; else val=low;
	}
      changed=wph_mspset(msp->i2c_client,0x0005,WPH_DSP,val,
			 0x000f,&msp->shadow);
    }
  else
    {
      int val=0;
      wph_mspget_shadow(msp->i2c_client,0x0005,WPH_DSP,&val,0x000f,
			&msp->shadow);
      voices[0]=(val==autogain)?low:val;
      voices[0]=high-voices[0];
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);  
  return changed;
}



int cb_avc_decay_time (snd_kmixer_element_t *element,int w_flag, 
				    int *voices)
{
  int changed=0;
  int low=0;
  int high=3;
  snd_card_msp3410D_t *msp=0;
  unsigned long flags;
  /*dprk(__FUNCTION__ ": element=0x%x write=%d voices[0]=%d\n",
    (int)element, w_flag,voices[0]);*/
  msp=msp_of_element(element);
  if(!msp) return -1;
  if(element!=msp->avc_decay_time)
    {
      /*eprk("element not found\n");*/
      /* this sometimes happend, but I don't know if it
	 is allowed to happen*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=voices[0];	/* between 0 and 3 */
      if(val<low||val>high)
	{
	  eprk("range error: val=%d\n",val);
	  if(val>high)val=high; else val=low;
	}
      val =(1<<val);
      changed=wph_mspset(msp->i2c_client,0x0029,WPH_DSP,val,
			 0x0f00,&msp->shadow);
    }
  else
    {
      int val=0;
      wph_mspget_shadow(msp->i2c_client,0x0029,WPH_DSP,&val,0x0f00,&msp->shadow);
      voices[0]=0;
      val>>=1;
      while (val!=0)
	{
	  voices[0]++;
	  val>>=1;
	}
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}



int cb_avc_switch(snd_kmixer_element_t *element, 
			int w_flag, unsigned int *bitmap)
{
  int changed=0;
  unsigned long flags;
  snd_card_msp3410D_t *msp=0;
  /*dprk(__FUNCTION__ ": w=%d value=%d\n",bitmap[0],w_flag);*/
  msp=msp_of_element(element);
  if(msp==0) return 0;
/*dprk("located switch on card %s\n",msp->id);*/
  if (element!=msp->sw1_avc_switch)
    {
      /*dprk(__FUNCTION__ ": element %x not recognised\n", 
	(int)element);*/
      /* this sometimes happens
	 but should not cause an error*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=(bitmap[0])?0x8:0x0;
      changed=wph_mspset(msp->i2c_client,0x0029,WPH_DSP,
			 val,0xf000,&msp->shadow);
    }
  else
    {
      int val;
      wph_mspget_shadow(msp->i2c_client,0x0029,WPH_DSP,&val,0xf000,&msp->shadow);
      bitmap[0]=(val==0x8);
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}

int cb_effect3D_highpass_auto(snd_kmixer_element_t *element, 
			int w_flag, unsigned int *bitmap)
{
  int changed=0;
  int autogain=8;
  unsigned long flags;
  snd_card_msp3410D_t *msp=0;
  /*dprk(__FUNCTION__ ": w=%d value=%d\n",bitmap[0],w_flag);*/
  msp=msp_of_element(element);
  if(msp==0) return 0;
/*dprk("located switch on card %s\n",msp->id);*/
  if (element!=msp->sw1_effect3D_highpass_auto)
    {
      /*dprk(__FUNCTION__ ": element %x not recognised\n", 
	(int)element);	*/
      /* this sometimes happens
				   but should not cause an error*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=snd_mixer_get_bit(bitmap,0);
      if(val)
	wph_msplockandset(msp->i2c_client,0x05,
			   WPH_DSP,autogain,0x000f,&msp->shadow);
      else
	wph_mspunlock(msp->i2c_client,0x05,WPH_DSP,
			   0x000f,&msp->shadow);
      changed=msp->highpass_auto_off==val; /* != is no mistake */
      msp->highpass_auto_off=!val;
    }
  else
    {
      snd_mixer_set_bit(bitmap,0,!msp->highpass_auto_off);
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}


int cb_effect3D_mode_switch(snd_kmixer_element_t *element, 
			int w_flag, unsigned int *bitmap)
{
  int changed=0;
  unsigned long flags;
  snd_card_msp3410D_t *msp=0;
  /*dprk("effect3D_mode_switch: w=%d value=%d\n",bitmap[0],w_flag);*/
  msp=msp_of_element(element);
  if(msp==0) return 0;
/*dprk("located switch on card %s\n",msp->id);*/
  if (element!=msp->sw1_effect3D_mode_switch)
    {
      /*dprk(__FUNCTION__ ": element %x not recognised\n", 
	(int)element);*/
      /* this sometimes happens
	 but should not cause an error*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=bitmap[0]?0x2:0x00;
      changed=wph_mspset(msp->i2c_client,0x05,WPH_DSP,val,0xf0,&msp->shadow);
      
    }
  else
    {
      int val;
      wph_mspget_shadow(msp->i2c_client,0x05,WPH_DSP,&val,0xf0,&msp->shadow);
      bitmap[0]=(val==0x2);
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}


int cb_equalizer_switch(snd_kmixer_element_t *element, 
			int w_flag, unsigned int *bitmap)
{
  int changed=0;
  unsigned long flags;
  snd_card_msp3410D_t *msp=0;
  /*dprk("equalizer_switch: w=%d value=%d\n",bitmap[0],w_flag);*/
  msp=msp_of_element(element);
  if(msp==0) return 0;
/*dprk("located switch on card %s\n",msp->id);*/
  if (element!=msp->sw1_equalizer_switch)
    {
      /*dprk("cb_equalizer_switch: element %x not recognised\n", 
	(int)element);*/
	/* this sometimes happens
				 but should not cause an error*/
      return 0;
    }
  spin_lock_irqsave(&msp->mixer_lock, flags);
  if(w_flag)
    {
      int val=bitmap[0]?0xff:0x00;
      changed=wph_mspset(msp->i2c_client,0x20,WPH_DSP,val,HIGH,&msp->shadow);
      
    }
  else
    {
      int val;
      wph_mspget_shadow(msp->i2c_client,0x20,WPH_DSP,&val,HIGH,&msp->shadow);
      bitmap[0]=((val&0xff)==0xff);
    }
  spin_unlock_irqrestore(&msp->mixer_lock, flags);
  return changed;
}





int delete_mixer(snd_card_msp3410D_t*msp)
{
  dprk("delete_mixer %s\n",msp->id);
  return 0;
}

int create_routes(snd_card_msp3410D_t * msp)
{
  snd_kmixer_t *mixer=msp->mixer;
  /*wphid;*/

  /*main volume control routed to loudspeaker output*/  
  if (snd_mixer_element_route_add(mixer, msp->sw1_volume_onoff_switch, 
				  msp->io_loudspeaker) < 0)
    goto __error;
 
  if (snd_mixer_element_route_add(mixer, msp->vol_main_volume, 
				  msp->sw1_volume_onoff_switch) < 0)
    goto __error;
 
  
  /*spatial effects routed to output main volume control*/

  /*wphid;*/
  if (snd_mixer_element_route_add
      (mixer, msp->sw1_effect3D_onoff_switch, 
       msp->vol_main_volume) < 0)
    goto __error;

  /*wphid;*/
  if (snd_mixer_element_route_add
      (mixer, msp->sw1_effect3D_mode_switch, 
       msp->sw1_effect3D_onoff_switch) < 0)
    goto __error;


  /*wphid;*/

  if (snd_mixer_element_route_add
      (mixer, msp->effect3D_strength, 
       msp->sw1_effect3D_onoff_switch) < 0)
    goto __error;
	
  /*complementary high pass routed to the 3D effect*/
  /*wphid;*/
  if (snd_mixer_element_route_add
      (mixer, msp->sw1_effect3D_highpass_auto, 
       msp->sw1_effect3D_onoff_switch) < 0)
    goto __error;

  if (snd_mixer_element_route_add
      (mixer, msp->effect3D_highpass_gain, 
       msp->sw1_effect3D_highpass_auto) < 0)
    goto __error;
	
  
  /*loudness routed to complementary high pass*/
  if (snd_mixer_element_route_add
      (mixer, msp->sw1_loudness_super_switch, 
       msp->effect3D_highpass_gain) < 0)
    goto __error;

  if (snd_mixer_element_route_add
      (mixer, msp->loudness_strength, 
       msp->sw1_loudness_super_switch) < 0)
    goto __error;
	
  /*main central accumulator routed to loudness (combines beeper and sound)*/
  /*if (snd_mixer_element_route_add(mixer,  msp->accu_central_in,
				  msp->loudness_strength) < 0)
				  goto __error;*/

  /*beeper injector routed to loudness*/
  if (snd_mixer_element_route_add(mixer, msp->sw1_beeper_onoff_switch,
				  msp->loudness_strength) < 0)
    goto __error;
  if (snd_mixer_element_route_add(mixer, msp->beeper_volume,
				  msp->sw1_beeper_onoff_switch) < 0)
    goto __error;
  if (snd_mixer_element_route_add(mixer, msp->beeper_freq,
				  msp->sw1_beeper_onoff_switch) < 0)
    goto __error;
  

  /*Equalizer routed to beeper injector*/

  if (snd_mixer_element_route_add(mixer, msp->sw1_equalizer_switch, 
				  msp->beeper_freq) < 0)
    goto __error;

  /*equalizer elements routed to Equalizer switch
   avc routed to equalizer elements*/
#ifdef EQ_ONE_GROUP
  if (snd_mixer_element_route_add
      (mixer, msp->tone_main_equalizer_tone_control, 
       msp->sw1_equalizer_switch) < 0)
    goto __error;
  /*wphid;*/ 
  if (snd_mixer_element_route_add
      (mixer, msp->sw1_avc_switch,
       msp->tone_main_equalizer_tone_control) < 0)
    goto __error;
#else 
 {
    int i;
    for (i=0; i<5;i++)
      {	
	/*wphid;*/ 
	if (snd_mixer_element_route_add
	(mixer, msp->tone_main_equalizer_tone_control[i], 
	msp->sw1_equalizer_switch) < 0)
	goto __error;
	/*wphid;*/ 
	if (snd_mixer_element_route_add
	    (mixer, msp->sw1_avc_switch,
	 msp->tone_main_equalizer_tone_control[i]) < 0)
	goto __error;
	}
  }
#endif
  /*wphid;*/
  /*bass treble tone control routed to Equalizer switch
   avc routed to bass and treble*/  
  /*wphid;*/
#ifdef EQ_ONE_GROUP
 if (snd_mixer_element_route_add
     (mixer, msp->tone_main_basstreble_tone_control, 
      msp->sw1_equalizer_switch) < 0)
   goto __error;
	/*wphid;*/
 if (snd_mixer_element_route_add
     (mixer, msp->sw1_avc_switch,
	     msp->tone_main_basstreble_tone_control) < 0)
   goto __error;
#else
  {
    int i;
    for (i=0; i<2;i++)
      {
	/*wphid;*/
	if (snd_mixer_element_route_add
	    (mixer, msp->tone_main_basstreble_tone_control[i], 
	     msp->sw1_equalizer_switch) < 0)
	  goto __error;
	/*wphid;*/
	if (snd_mixer_element_route_add
	    (mixer, msp->sw1_avc_switch,
	     msp->tone_main_basstreble_tone_control[i]) < 0)
	  goto __error;
      }
  }
#endif
  
  
  /*avc routed to equalizer and bass treble (see above)*/
  /*wphid;*/
  if (snd_mixer_element_route_add
      (mixer, msp->avc_decay_time, 
       msp->sw1_avc_switch) < 0)
    goto __error;
  
/*loudspeaker channel matrix routed to avc*/

  
  /*sound input routed to loudspeaker channel matrix*/

  /*wphid;*/
  if (snd_mixer_element_route_add
      (mixer, msp->io_video_in,msp->avc_decay_time) < 0)
    goto __error;
  return 0;
  
 __error:
  /*wphid;*/
  return -ENOMEM;
}

#if 0
static int snd_sb16mixer_cd_input_switch(snd_kmixer_element_t *element, int w_flag, unsigned int *prsw)
{
  dprk("prsw[0]=0x%x\n",prsw[0]);
  /*snd_mixer_set_bit(prsw, 0, 0);*/
  if(!w_flag)
    {	
      static int called=0;
      called++;
      snd_mixer_set_bit(prsw, 0, called%2);
      snd_mixer_set_bit(prsw, 1, (called/2)%2);
      snd_mixer_set_bit(prsw, 2, ((called/4)%2));
      snd_mixer_set_bit(prsw, 3, (called/8)%2);
    }
  else
    {
      snd_mixer_get_bit(prsw, 0);
      snd_mixer_get_bit(prsw, 1);
      snd_mixer_get_bit(prsw, 2);
      snd_mixer_get_bit(prsw, 3);
    }
  /*snd_mixer_element_write(element->card->mixer, element);*/
  /*snd_mixer_element_change((void*)element->private_data,element);*/
  return 1;
}
#endif


int create_mixer(snd_card_msp3410D_t * msp, int device)
{
  snd_kmixer_t *mixer;
  snd_kmixer_group_t *group,*group_equalizer;
#ifndef EQ_ONE_GROUP
  snd_kmixer_group_t*group_basstreble;
#endif
  int err;
  static struct snd_mixer_element_volume1_range main_volume_range[2] = {
    {0, 100, -114, 0},
    {0, 100, -114, 0}
  };
#ifdef EQ_ONE_GROUP
  static struct snd_mixer_element_volume1_range eq_range[5] = {
    {-50,50, -12, 12},
    {-50,50, -12, 12},
    {-50,50, -12, 12},
    {-50,50, -12, 12},
    {-50,50, -12, 12},
  };
  static struct snd_mixer_element_tone_control1_info tone_control =
  {
    SND_MIXER_TC1_BASS | SND_MIXER_TC1_TREBLE,
    -50, 50, -1400, 1400,
    -50, 50, -12, 12
  };
  
#else
  static struct snd_mixer_element_volume1_range eq_range[1] = {
    {-50,50, -12, 12},
  };
#endif
  static struct snd_mixer_element_volume1_range effect3D_range[1] = {
    {-50,50, -120, 120},
  };
  static struct snd_mixer_element_volume1_range highpass_gain_range[1] = {
    {0,6, 0, 6},
  };
  static struct snd_mixer_element_volume1_range loudness_range[1] = {
    {0,100, 0, 17},
  };
  static struct snd_mixer_element_volume1_range beeper_volume_range[1] = {
    {0,100, -114,0},
  };
  static struct snd_mixer_element_volume1_range beeper_freq_range[1] = {
    {0, 100, 0, 100},
  };
  static struct snd_mixer_element_volume1_range avc_time_range[1] = {
    {0, 3, 0, 3},
  };
  wphid;
  snd_debug_check(msp == NULL, -EINVAL);
  /*create the mixer*/
  if ((err = snd_mixer_new(msp->card, "Msp3410D mixer", 
			   device, &mixer)) < 0)
    return err;
  strcpy(mixer->name, mixer->id);


  wphid;
  /*loudpspeaker output*/
  if ((msp->io_loudspeaker =
       snd_mixer_lib_io_stereo(mixer,"Loudspeakers", 
			       0, SND_MIXER_ETYPE_OUTPUT, 0)) == NULL)
    goto __error;

  /*main volume control routed to loudspeaker output*/  
  if ((group = snd_mixer_lib_group(mixer, SND_MIXER_OUT_MASTER, 0)) == NULL)
    goto __error;
  wphid;
  if ((msp->sw1_volume_onoff_switch = 
       snd_mixer_lib_sw1(mixer,"Vol On/Off", 0, 1, 
			 cb_volume_onoff_switch, msp)) == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add (mixer, group, 
       msp->sw1_volume_onoff_switch) < 0)
    goto __error;
  if ((msp->vol_main_volume = snd_mixer_lib_volume1(mixer, "Main Volume", 
						    0, 2, main_volume_range, 
						    cb_main_volume, msp)) == NULL)
    goto __error;
  if (snd_mixer_group_element_add(mixer, group, msp->vol_main_volume) < 0)
    goto __error;

  /*main output accumulator, which merges different signals into one*/
  /*if ((msp->accu_main_out = 
       snd_mixer_lib_accu1(mixer,SND_MIXER_ELEMENT_OUTPUT_ACCU, 0, 0
			   )) == NULL)
    goto __error;
  */
  
  /*loudness routed to output accumulator*/

  wphid;
  if ((group = snd_mixer_lib_group(mixer,"Bass Boost", 0)) == NULL)
    goto __error;
#if 0
  /*wphtest*/
  {	
    static snd_mixer_voice_t stereo_voices[2] =
    {
      {SND_MIXER_VOICE_LEFT, 0},
      {SND_MIXER_VOICE_RIGHT, 0}
    };
    snd_kmixer_element_t *test=0;
    if ((test = snd_mixer_lib_sw3(mixer, "CD Input Switch", 0, SND_MIXER_SWITCH3_FULL_FEATURED, 
				  2, stereo_voices, snd_sb16mixer_cd_input_switch, msp)) == NULL)
      goto __error;
    if (snd_mixer_group_element_add (mixer, group, 
				     test) < 0)
      goto __error;
    if (snd_mixer_element_route_add(mixer, test,msp->vol_main_volume ) < 0)
      goto __error;
  }
#endif

  wphid;
  if ((msp->sw1_loudness_onoff_switch = 
       snd_mixer_lib_sw1(mixer,"Loudness On/Off", 0, 1, 
			 cb_loudness_onoff_switch, msp)) == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add (mixer, group, 
       msp->sw1_loudness_onoff_switch) < 0)
    goto __error;
  wphid;
  if ((msp->sw1_loudness_super_switch = 
       snd_mixer_lib_sw1(mixer,"Super boost", 0, 1, 
			 cb_loudness_super_switch, msp)) == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add (mixer, group, 
       msp->sw1_loudness_super_switch) < 0)
    goto __error;

  wphid;
  if ((msp->loudness_strength = 
       snd_mixer_lib_volume1(mixer,"Loudness strength", 0, 1, loudness_range, 
			     cb_loudness_strength, msp)) == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add
      (mixer, group,msp->loudness_strength) < 0)
    goto __error;
  wphid;
	



  /*spatial effects routed to output main output accu*/

  wphid;
  if ((group = snd_mixer_lib_group(mixer, SND_MIXER_GRP_EFFECT_3D, 0)) == NULL)
    goto __error;
  wphid;
  if ((msp->sw1_effect3D_onoff_switch = 
       snd_mixer_lib_sw1(mixer,"3D On/Off", 0, 1, 
			 cb_effect3D_onoff_switch, msp)) == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add (mixer, group, 
       msp->sw1_effect3D_onoff_switch) < 0)
    goto __error;
  wphid;
  if ((msp->sw1_effect3D_mode_switch = 
       snd_mixer_lib_sw1(mixer,"SBE+PSE/SBE", 0, 1, 
			 cb_effect3D_mode_switch, msp)) == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add (mixer, group, 
       msp->sw1_effect3D_mode_switch) < 0)
    goto __error;

  wphid;
  if ((msp->effect3D_strength = 
       snd_mixer_lib_volume1(mixer,"3D strength", 0, 1, effect3D_range, 
			     cb_effect3D_strength, msp)) == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add
      (mixer, group,msp->effect3D_strength) < 0)
    goto __error;
	

  /*complementary high pass routed to the 3D effect*/
  /*we use the same group as above*/
  
  wphid;
  if ((msp->sw1_effect3D_highpass_auto = 
       snd_mixer_lib_sw1(mixer,"High Pass Auto Gain", 0, 1, 
			 cb_effect3D_highpass_auto, msp)) == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add (mixer, group, 
       msp->sw1_effect3D_highpass_auto) < 0)
    goto __error;

  wphid;
  if ((msp->effect3D_highpass_gain = 
       snd_mixer_lib_volume1(mixer,"High Pass Gain", 0, 1, 
			     highpass_gain_range, 
			     cb_effect3D_highpass_gain, msp)) == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add
      (mixer, group,msp->effect3D_highpass_gain) < 0)
    goto __error;
  wphid;
	


  wphid;
  /*Beep injection */  
  if ((group = snd_mixer_lib_group(mixer,"Beep inject", 1))
      == NULL)
    goto __error;
  wphid;
  if ((msp->sw1_beeper_onoff_switch = 
       snd_mixer_lib_sw1(mixer,"Beeper On/Off", 0, 1, 
			 cb_beeper_onoff_switch, msp)) == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add (mixer, group, 
       msp->sw1_beeper_onoff_switch) < 0)

  wphid;
  if ((msp->beeper_volume = 
       snd_mixer_lib_volume1(mixer,"Beeper volume", 
			     0, 1, beeper_volume_range, 
				   cb_beeper_volume, msp))
      == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add
      (mixer, group,msp->beeper_volume) < 0)
    goto __error;

  wphid;
  if ((msp->beeper_freq = 
       snd_mixer_lib_volume1(mixer,"Beeper Frequency", 
			     0, 1, beeper_freq_range, 
				   cb_beeper_freq, msp))
      == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add
      (mixer, group,msp->beeper_freq) < 0)
    goto __error;


  wphid;
  /*Equalizer switch routed to central accumulator*/
  if ((group = snd_mixer_lib_group(mixer, SND_MIXER_GRP_EQUALIZER, 0)) == NULL)
    goto __error;
   group_equalizer=group;
  if ((msp->sw1_equalizer_switch = 
       snd_mixer_lib_sw1(mixer,"Equalizer Switch", 0, 1, 
			 cb_equalizer_switch, msp)) == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add(mixer, group, msp->sw1_equalizer_switch) < 0)
    goto __error;
  wphid;
  /*equalizer tone control routed to Equalizer switch*/
#ifdef EQ_ONE_GROUP 			/* wph */
  if ((msp->tone_main_equalizer_tone_control = 
       snd_mixer_lib_volume1(mixer,SND_MIXER_GRP_EQUALIZER, 
			     0, 5, eq_range,
			     cb_main_equalizer_tone_control, msp))
      == NULL)
    goto __error;
  /*  msp->tone_main_equalizer_tone_control->type= 
      SND_MIXER_ETYPE_EQUALIZER1 ;*/
  wphid;
  if (snd_mixer_group_element_add
      (mixer, group, 
       msp->tone_main_equalizer_tone_control) < 0)
#else
 {
    int i;
    for (i=0; i<5;i++)
      {
       /* we create a separate control for each band because
	most mixers have an option for locking controls together;
       if the controls are locked the equalizer would not function properly*/
	wphid;
	if ((msp->tone_main_equalizer_tone_control[i] = 
	     snd_mixer_lib_volume1(mixer,SND_MIXER_GRP_EQUALIZER, 
				   i+2, 1, eq_range,
				   cb_main_basstreble_tone_control, msp))
	    == NULL)
	  goto __error;
	wphid;
	if (snd_mixer_group_element_add
	     (mixer, group, 
	      msp->tone_main_equalizer_tone_control[i]) < 0)
	  goto __error;
      }
  }
#endif
  wphid;
  /*bass treble tone control routed to Equalizer switch */  
#ifdef EQ_ONE_GROUP
  if ((group = snd_mixer_lib_group(mixer,SND_MIXER_GRP_BASS, 1))
      == NULL)
    goto __error;
  wphid;
  if ((msp->tone_main_basstreble_tone_control= 
       snd_mixer_lib_tone_control1(mixer,
			     SND_MIXER_ELEMENT_TONE_CONTROL, 
			     0, &tone_control, 
			     cb_main_basstreble_tone_control, msp))
      == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add
      (mixer, group, 
       msp->tone_main_basstreble_tone_control) < 0)
    goto __error;
  if ((group = snd_mixer_lib_group(mixer,SND_MIXER_GRP_TREBLE, 1))
      == NULL)
    goto __error;
  if (snd_mixer_group_element_add
      (mixer, group, 
       msp->tone_main_basstreble_tone_control) < 0)
    goto __error;
#else
  if ((group = snd_mixer_lib_group(mixer,"Bass/Treble", 1))
      == NULL)
    goto __error;
  group_basstreble=group;
  wphid;
  {
    int i;
    for (i=0; i<2;i++)
      {
       /* we create a separate control for bass/treble because
	most mixers have an option for locking controls together;
       if the controls are locked the control would not function properly*/
	if ((msp->tone_main_basstreble_tone_control[i] = 
	     snd_mixer_lib_volume1(mixer,
				   i?SND_MIXER_GRP_TREBLE:SND_MIXER_GRP_BASS, 
				   i, 1, eq_range, 
				   cb_main_equalizer_tone_control, msp))
	    == NULL)
	  goto __error;
	wphid;
	if (snd_mixer_group_element_add
	    (mixer, group, 
	     msp->tone_main_basstreble_tone_control[i]) < 0)
	  goto __error;
      }
  }
#endif

  
  /*avc routed to main tone control*/
  if ((group = snd_mixer_lib_group(mixer,"AVC", 0)) == NULL)
    goto __error;
  wphid;
  if ((msp->sw1_avc_switch = 
       snd_mixer_lib_sw1(mixer,"AVC on/off", 0, 1, 
			 cb_avc_switch, msp)) == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add (mixer, group, 
       msp->sw1_avc_switch) < 0)
    goto __error;
  wphid;
  if ((msp->avc_decay_time = 
       snd_mixer_lib_volume1(mixer,"Decay time", 0, 1, 
			     avc_time_range, 
			     cb_avc_decay_time, msp)) == NULL)
    goto __error;
  wphid;
  if (snd_mixer_group_element_add
      (mixer, group,msp->avc_decay_time) < 0)
    goto __error;
  wphid;

  
/*loudspeaker channel matrix routed to avc*/
  /*not yet implemented*/
  
  /*sound input routed to loudspeaker channel matrix*/
  wphid;
  if ((msp->io_video_in =
       snd_mixer_lib_io_stereo(mixer, SND_MIXER_IN_VIDEO, 
			       0, SND_MIXER_ETYPE_INPUT, 0)) == NULL)
    goto __error;
  wphid;
  mixer->private_data = msp;
  if(create_routes(msp)<0) 
    goto __error;
  return 0;
  
 __error:
  wphid;
  /*temporarily off*/
  /* snd_device_free(msp->card, mixer);
  */
  return -ENOMEM;
}







module_init(alsa_msp3410D_module_init);
module_exit(alsa_msp3410D_module_exit);



MODULE_DESCRIPTION("\n\
Driver: msp3410D soundcard (/dev/null)\n\
ALWAYS\n\
");

#if 0				/* wph: turn off for now */
MODULE_PARM(snd_index, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(snd_index, "Index value for msp3410D soundcard.");
MODULE_PARM(snd_id, "1-" __MODULE_STRING(SND_CARDS) "s");
MODULE_PARM_DESC(snd_id, "ID string for msp4310 soundcard.");
MODULE_PARM(snd_enable, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(snd_enable, "Enable this msp3410D soundcard. [BOOL]");
#endif


/*
 * Local variables:
 *  compile-command: "sudo make &&sudo make install && sudo depmod -a"
 * End:
 */
