/*
  
  File msputil.c -  utility routines for msp34xx

  
  Copyright (c) 1997-2000 Gerd Knorr  
  Copyright (c) 2000      Wilfried Philips  
  
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330 
  Boston, MA  02111-1307 USA 
*/

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/videodev.h>
#include <linux/i2c.h>
#include "msputil.h"
#include "mspdata.h"

extern int debug;

inline int protmask(int lock)
{
  int ret=(lock&1)?0xf:0;
  ret+= (lock&2)  ?0xf0:0;
  ret+= (lock&4)  ?0xf00:0;
  ret+= (lock&8)  ?0xf000:0;
  return ret;
}

inline int unprotmask(int mask)
{
  int ret=(mask&0xf) ?1:0;
  ret+= (mask&0xf0)  ?2:0;
  ret+= (mask&0xf00) ?4:0;
  ret+= (mask&0xf000)?8:0;
  return ret;
}


int snprintf(char*p,int size, char *fmt,...)
{
        va_list args;
        int len;
        char sbuffer[512];
        va_start(args, fmt);
        len = vsprintf(sbuffer, fmt, args);
        va_end(args);
	if(len>sizeof(sbuffer))
	  {
	    printk("wrote outside of buffer\n");
	  }
	if(size<=0)
	  return len;
	strncpy(p,sbuffer,(size<len)?size:len);
	if(len>size)
	  p[size-1]=0;
	if(len>size){
	  len=size;
	  sbuffer[len-1]='0';
	}
        return len;
}

int msp3400c_reset(struct i2c_client *client)
{
        static char reset_off[3] = { 0x00, 0x80, 0x00 };
        static char reset_on[3]  = { 0x00, 0x00, 0x00 };

        i2c_master_send(client,reset_off,3);  /* XXX ignore errors here */
        if (3 != i2c_master_send(client,reset_on, 3)) {
		printk(KERN_ERR "msp3400: chip reset failed, penguin on i2c bus?\n");
                return -1;
	}
        return 0;
}

int			
msp3400c_read(struct i2c_client *client, int dev, int addr)
{
	int err;

        unsigned char write[3];
        unsigned char read[2];
        struct i2c_msg msgs[2] = {
                { client->addr, 0,        3, write },
                { client->addr, I2C_M_RD, 2, read  }
        };
        write[0] = dev+1;
        write[1] = addr >> 8;
        write[2] = addr & 0xff;

	for (err = 0; err < 3;) {
		if (2 == i2c_transfer(client->adapter,msgs,2))
			break;
		err++;
		printk(KERN_WARNING "msp34xx: I/O error #%d (read 0x%02x/0x%02x)\n",
		       err, dev, addr);
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(HZ/10);
	}
	if (3 == err) {
		printk(KERN_WARNING "msp34xx: giving up, reseting chip. Sound will go off, sorry folks :-|\n");
		msp3400c_reset(client);
		return -1;
	}
        return read[0] << 8 | read[1];
}


int			
msp3400c_write(struct i2c_client *client, int dev, int addr, int val)
{
	int err;
        unsigned char buffer[5];

        buffer[0] = dev;
        buffer[1] = addr >> 8;
        buffer[2] = addr &  0xff;
        buffer[3] = val  >> 8;
        buffer[4] = val  &  0xff;

	for (err = 0; err < 3;) {
		if (5 == i2c_master_send(client, buffer, 5))
			break;
		err++;
		printk(KERN_WARNING "msp34xx: I/O error #%d (write 0x%02x/0x%02x)\n",
		       err, dev, addr);
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(HZ/10);
	}
	if (3 == err) {
		printk(KERN_WARNING "msp34xx: giving up, reseting chip. Sound will go off, sorry folks :-|\n");
		msp3400c_reset(client);
		return -1;
	}
	return 0;
}




int maskandshiftright(int val, int mask)
{
  if(mask==0)
    {
      eprk(__FUNCTION__ "cannot deal with mask=0\n");
      return val;
    }
  val &=mask;
  if(!(mask&0xff))
    {
      mask >>=8;
      val >>=8;
    }
  while(!(mask&1))
    {
      mask >>=1;
      val >>=1;
    }
  return val;
}

int shiftleftandmask(int val, int mask)
{
  int m=mask;
  if(mask==0)
    {
      eprk(__FUNCTION__ "cannot deal with mask=0\n");
      return val;
    }
  if(!(mask&0xff))
    {
      mask >>=8;
      val <<=8;
    }
  while(!(mask&1))
    {
      mask >>=1;
      val <<=1;
    }
  val &=m;
  return val;
}





void dump_shadow(msp_shadow *s)
{
  int i;
  for(i=0;i<MSP_DSP_NUMREG;i++)
    {
      printk("0x%4.4x=0x%4.4x %s\n",i,s->dsp_reg[i],s->dsp_flags[i].valid?"(valid)":"");
    }
}





/*read a register and as a side effect save its value in the shadow register*/
int  wph_mspget(struct i2c_client *client,int addr,int dev,int *val, int mask,msp_shadow *s)
{
  int t=0;
  int prot=0;
  int regval=0,shadowval=0;
  switch(dev)
    {
    case WPH_DSP:
      t=I2C_MSP3400C_DFP;
      break;
    case WPH_DEM:
      t=I2C_MSP3400C_DEM;
      break;
    default:
      return -1;
    }
  *val=regval=msp3400c_read(client,t,addr);
  /*dprk("msp read: 0x%4.4x=0x%4.4x\n",addr,*val);*/
  if (*val == -1) return -1;
  if(s&&(dev==WPH_DSP))
    {
      if(addr<0||addr>=MSP_DSP_NUMREG)
	{
	  eprk(__FUNCTION__ ": no shadow register for address 0x%4.4x\n",addr);
	}
      else
	{
	  prot= protmask(s->dsp_flags[addr].lock);
	  shadowval=s->dsp_reg[addr];
	  *val=(regval&~prot)|(shadowval&prot);
	  s->dsp_reg[addr]=*val;
	  s->dsp_flags[addr].valid=1;
	}
    }
  dprk(__FUNCTION__ 
       ":0x%4.4x(0x%4.4x) reg=0x%4.4x "
       "shadow=0x%4.4x(0x%4.4x)->0x%4.4x\n",addr,mask,
       s?regval:-1,
       shadowval,prot,*val);
  *val=maskandshiftright(*val,mask);
  return 0;
}

/*read a shadow register; if the shadow register is not available
  read  the real register*/
int  wph_mspget_shadow(struct i2c_client*client, int addr,int dev,int *val, int mask,msp_shadow*s)
{
  int v=0;
  switch(dev)
    {
    case WPH_DSP:
      if(!s) break;
      if(addr<0||addr>=MSP_DSP_NUMREG)
	{
	  /*eprk(__FUNCTION__ ": no shadow register for address 0x%4.4x\n",addr);*/
	  break;
	}
      v=s->dsp_flags[addr].valid;
      *val=s->dsp_reg[addr]; 
      break;
    case WPH_DEM:
      break;
    default:
      break;
    }
  if(!v) 
    {
      return wph_mspget(client,addr,dev,val,mask,s);
    }
  /*dprk("msp read: 0x%4.4x=0x%4.4x\n",addr,*val);*/
  if (*val == -1) return -1;
  *val=maskandshiftright(*val,mask);
  return 0;
}






/*write both the shadow register and the real register*/
/*return: <0 -> error
          =0 -> no error, but new value is same as old
          =1 -> no error; ne value is different than old
*/
static int  wph_mspset_sub(struct i2c_client *client,int addr,int dev,
			   int val,int mask,msp_shadow *s,int setshadow)
{
  int regval=0;			/* read from chip */
  int shadowval=0;		/* read from shadow register */
  int newshadowval=0,newregval=0;
  int t=0;
  int ret=1;
  int prot=0;
  switch(dev)
    {
    case WPH_DSP:
      t=I2C_MSP3400C_DFP;
      if(s&&(dev==WPH_DSP))
	{
	  if(addr<0||addr>=MSP_DSP_NUMREG)
	    {
	      eprk(__FUNCTION__ ": no shadow register for address"
		   " 0x%4.4x\n",addr);
	    }
	  else
	    {
	      prot= protmask(s->dsp_flags[addr].lock);
	      shadowval=s->dsp_reg[addr];
#if 0
	      if(!s->dsp_flags[addr].valid)
		{
		  mask &= ~prot;
		  prot=0;
		}
#endif
	    }
	}
      break;
    case WPH_DEM:
      t=I2C_MSP3400C_DEM;
      break;
    default:
      return -1;
    }
  /*mask: 0 indicates that bit is not written to shadow nor to register
    mask&prot: 1 indicates that bit is written only to shadow
    mask&~prot: 1 indicates that bit is written to shadow and register
  */
  mask=mask&0xffff;
  if(mask!=0xffff)
    regval=msp3400c_read(client,t,addr);
  val=shiftleftandmask(val,mask);
  
  /*dprk("msp write: 0x%4.4x=0x%4.4x ->",addr,val);*/
  newshadowval=(shadowval&prot&~mask)|(regval&~prot&~mask)|(val&mask);
  newregval=setshadow?
	(regval&(~mask|prot))|(val&mask&~prot)	/* we may only write 
						   non-protected bits */
    :((regval&~mask)|(val&mask)); /* if we bypass shadow register 
				   we may write anything we want*/
  /*dprk("wph_mspset: 0x%4.4x mask=%x\n",val,mask);*/

  dprk(__FUNCTION__ 
       ":0x%4.4x(0x%4.4x):val=%4.4x: reg=0x%4.4x->0x%4.4x "
       "shadow=0x%4.4x(0x%4.4x)->0x%4.4x\n",addr,mask,val,
       regval,newregval,
       s?shadowval:-1,prot,setshadow?newshadowval:(s?shadowval:-1));

  ret=msp3400c_write(client,t ,addr,newregval);
  if(s&&(dev==WPH_DSP))
    {
      if(addr<0||addr>=MSP_DSP_NUMREG)
	{
	  eprk(__FUNCTION__ ": no shadow register for address 0x%4.4x\n",addr);
	}
      else
	{
	  ret=(!s->dsp_flags[addr].valid||
	       (shadowval&mask)!=(newshadowval&mask));
	  if(setshadow)
	    {
	      s->dsp_flags[addr].valid=1;
	      s->dsp_reg[addr]=newshadowval;
	    }
	}
    }
  return ret;
}


int  wph_mspset(struct i2c_client *client,int addr,int dev,int val,
		int mask,msp_shadow *s)
{
  return wph_mspset_sub(client,addr,dev,val,mask,s,1);
}


/*
  protect bits of msp regsiter; a protected bit will never be written
  note: bits can be protected/unprotected in groups op 4 bit only;
 -> valid values of mask are 0x.... with . equal to 0 or f
 lock can be 1 or 0
 */
int  wph_msplock(int addr,int dev,int lock,int mask,msp_shadow *s)
{
  int prot=0,newprot=0;
  switch(dev)
    {
    case WPH_DSP:
      if(s&&(dev==WPH_DSP))
	{
	  if(addr<0||addr>=MSP_DSP_NUMREG)
	    {
	      eprk(__FUNCTION__ ": no shadow register for address"
		   " 0x%4.4x\n",addr);
	    }
	  else
	    {
	      lock= (lock)?mask:0;
	      prot=protmask(s->dsp_flags[addr].lock);
	      newprot=((prot&~mask)|lock);
	      s->dsp_flags[addr].lock=unprotmask(newprot);
	      dprk(__FUNCTION__ 
		   ":0x%4.4x(0x%4.4x):val=%4.4x: "
		   "lock=0x%4.4x->0x%4.4x(%4.4x)\n",
		   addr,mask,lock,prot,newprot,
		   s->dsp_flags[addr].lock);
	    }
	}
      break;
    case WPH_DEM:
      break;
    default:
      return -1;
    }
  return 0;
}

int  wph_msplockandset(struct i2c_client *client,int addr,int
			   dev,int val,int mask,msp_shadow *s)
{
  wph_msplock(addr,dev,1,mask,s);
  return 	  
    wph_mspset_sub(client,addr,dev,val,mask,s,0);
}

int  wph_mspunlock(struct i2c_client *client,int addr,int
			   dev,int mask,msp_shadow *s)
{
  int val=0;
  wph_msplock(addr,dev,0,mask,s);
  wph_mspget_shadow(client,addr,dev,&val,mask,s); 
  return 	  
    wph_mspset(client,addr,dev,val,mask,s);
}


#if 0
/*write only to the shadow register*/
int  wph_mspset_shadowonly(struct i2c_client *client,int addr,int
			   dev,int val,int mask,msp_shadow *s)
{
  int oldval=0;
  int t=0;
  int ret=1;
  mask=mask&0xffff;
  if(mask!=0xffff)
    ret=wph_mspget_shadow(client, addr,dev,&oldval, 0xffff,s);
  val=shiftleftandmask(val,mask);
  /* dprk("msp write: 0x%4.4x=0x%4.4x ->",addr,val);*/
  val=(oldval&~mask)|(val&mask);
  /*dprk("wph_mspset: 0x%4.4x mask=%x\n",val,mask);*/
  if(s&&(dev==WPH_DSP))
    {
      if(addr<0||addr>=MSP_DSP_NUMREG)
	{
	  eprk(__FUNCTION__ ": no shadow register for address 0x%4.4x\n",addr);
	  ret=-1;
	}
      else
	{
	  ret=(!s->dsp_flags[addr].valid||(s->dsp_reg[addr]&mask)!=(val&mask));
	  s->dsp_flags[addr].valid=1;
	  s->dsp_reg[addr]=val;
	}
    }
  return ret;
}
#endif


/*
 * Local variables:
 *  compile-command: "sudo make &&sudo make install && sudo depmod -a"
 * End:
 */
