/*
  
  File wphproc.c - /proc routines for msp34xx

  
  Copyright (c) 2000 W. Philips
  
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330 
  Boston, MA  02111-1307 USA 
*/

#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/proc_fs.h>
#include <linux/videodev.h>
#include <linux/i2c.h>
#include "mspdata.h"
#include "msputil.h"


#define procname "msp"
int debug=1;

#define MAXMSPS 4

static struct wphprocdata
{
  int nmsps;
  struct proc_dir_entry *proc_dir_entries[MAXMSPS];
} private;
char mymess[4096*4];




int procfile_read(char *page, 
                  char **start, 
                  off_t off, 
                  int count, int *eof,
                  void *data)
{
  int len=0;  /* The number of bytes actually used */
  static int curlen=0;
  int clen;
  int val;
  char *name;
  static int called=0;
  int rem=sizeof(mymess);
  int i,j,addr,hl;
  char *p=mymess;
  struct i2c_client *client;
  struct msp_val_name_struct*vn=0;
  int mask=0;
  char *valname=0;
  *start=mymess;
  *eof=1;
  called++;
  printk("procfile_read nmsps=%d\n",private.nmsps);
  if (!private.nmsps)
    return 0;
  if (off>0) 
    {
      *start=mymess+off;
      len=curlen-off;
      if(len<0) len=0;
      *eof=(len<count);
      return len;
    }
  for(j=0;j<private.nmsps;j++)
     {
      client=private.proc_dir_entries[j]->data;
      printk("procfile_read client=0x%x nmsps=%d\n",(int)client,private.nmsps);
      for(i=0;i<private.nmsps;i++)
	{
	  clen=snprintf(p,rem,"Demodulator:\n");
	  p+=clen;rem-=clen;if(rem<=0) goto end;
	  for(i=0;msp_dem_addrs[i].addr>=0;i++)
	    {
	      if(msp_dem_addrs[i].t!='r')
		continue;
	      addr=msp_dem_addrs[i].addr;
	      name=msp_dem_addrs[i].name;
	      val=0;
	      val=msp3400c_read(client,I2C_MSP3400C_DEM,addr);
	      valname=find_name_of_val(vn,val,mask);
	      clen=snprintf(p,rem,"DEM%4.4x=0x%4.4x %s=%s\n",addr,val,name,
			    valname);
	      p+=clen;rem-=clen;if(rem<=0) goto end;
	    }

	  clen=snprintf(p,rem,"DSP:\n");
	  p+=clen;rem-=clen;if(rem<=0) goto end;
	  for(i=0;msp_dsp_addrs[i].addr>=0;i++)
	    {
	      addr=msp_dsp_addrs[i].addr;
	      name=msp_dsp_addrs[i].name;
	      mask=msp_dsp_addrs[i].mask;
	      vn=msp_dsp_addrs[i].vn;
	      hl=msp_dsp_addrs[i].hl;
	      val=0;
	      val=msp3400c_read(client,I2C_MSP3400C_DFP,addr);
	      valname=find_name_of_val(vn,val,mask);
	      switch(hl)
		{
		case 'H':
		  clen=snprintf(p,rem,"DSP%4.4xH=0x%-4.2x %s=%s\n",addr,val>>8,
				name,valname);
		  break;
		case 'L':
		  clen=snprintf(p,rem,"DSP%4.4xL=0x%-4.2x %s=%s\n",addr,
				val&0xff,name,valname);
		  break;
		case 0:
		  clen=snprintf(p,rem,"DSP%4.4x=0x%4.4x %s=%s\n",addr,val,
				name,valname);
		  break;
		}
	      p+=clen;rem-=clen;if(rem<=0) goto end;
	    }
	  
	}
    }
 end:
  curlen=p-mymess;
  len=curlen;
  *eof=(len<count);
  return len;
}









enum {ISLOW,ISHIGH,ISWIDE} ADDRTYPE;




 int procfile_write(struct file *file, const char *buffer,
                           unsigned long count, void *data)
{
  struct i2c_client *client=data;
  const int str_size=256;
  char *str=kmalloc(str_size,GFP_KERNEL);
  char *hex=kmalloc(str_size,GFP_KERNEL);
  const char *dsp="DSP";
  const char *dem="DEM";
  char *found=0;
  char *found1=0;
  char *found2=0;
  int n=str_size-1;
  int type=0;
  int addr=0;
  int mask=0;
  int val=0;
  strncpy(hex,"0x",2);
  if(count<n) n=count;
  strncpy(str,buffer,n);
  str[n]=0;
  found1=strstr(str,dsp);
  found2=strstr(str,dem);
  if((found1-found2)<0)
    {type=WPH_DSP;found=found1+sizeof(dsp)-1;}
  else  if((found2-found1)<0)
    {type=WPH_DEM;found=found2+sizeof(dem)-1;}
  else
    type=WPH_UNSET;
  if(type!=WPH_UNSET)
    {
      if(found-str+5>=n)	/* need 4 digits and 1 of 'H' 'L' or '=' */
	type=WPH_UNSET;
    }
  else found=str+n;
  if(type!=WPH_UNSET)
    {
      strncpy(hex+2,found,4);	/* 4 hex digits */
      hex[6]=0;
      addr=simple_strtoul(hex,0,0);found+=4;
      mask=0xffff;
      if(type==WPH_DSP)
	{
	  if(found[0]=='H') mask=0xff00;
	  else if(found[0]=='L') mask=0x00ff;		  
	}
      found++;
      if(mask!=0xffff)found++; /* now found points after '=' */
      if (found-str>=n|| found[-1]!='=')
	type=WPH_UNSET;
    }
  if(type!=WPH_UNSET)
    {
      if((found-str+4>=n)
	 ||((found-str+6>=n)&&mask==0xffff)
	 ||found[0]!='0'
	 ||found[1]!='x')
	type=WPH_UNSET;
      else 
	{
	  int l=(mask==0xffff)?4:2;
	  found+=2;
	  strncpy(hex+2,found,l);	/*  4 or 2 hex digits */
	  hex[2+l]=0;
	  val=simple_strtoul(hex,0,0);found+=l;
	}
    }
  /* now skip to end of line, which is (dangerously) assumed to 
     fall within current str
   */
  { 
    char *f;
    f=strchr(found,'\n');
    if(!f) type=WPH_UNSET;
    else found=f+1;
  }
  if(found-str<count) count=found-str;
  if(count>=n)count=n;
  if(type!=WPH_UNSET)
    printk("wphproc: set: %s%4.4x%c=%4.4x\n",
	 (type==WPH_DEM)?dem:dsp,addr,
	 (mask==0xffff)?'W': ((mask==0x00ff)?'L':'H'),
	 val);
  wph_mspset(client, addr,type, val,mask,0);
  kfree(hex);
  kfree(str);
  return count;
}


struct proc_dir_entry* create_procfile(int i,struct i2c_client*c)
{
  struct proc_dir_entry *d;
  char name[20];
  int len=0;
  dprk("init_wph_proc\n");
  len=1+sprintf(name,"%s%d",procname,i);
/* File mode - this is a regular 
 * file which can be read and written 
 * by its * owner, its group, and everybody
 * else*/
  d=create_proc_entry(name,S_IFREG|S_IRUGO|S_IWUGO,&proc_root);
  if(!d) 
    {
      eprk("create_proc_entry failed\n");
      return 0; 
    }
  d->size=80;   /* The size of the file reported by ls.*/
  d->read_proc=procfile_read;
  d->write_proc=procfile_write;
  d->data=c;
  return d;
}


void delete_procfile(struct proc_dir_entry*d)
{
  
  dprk("removing proc entry %s\n", d->name);
  remove_proc_entry(d->name,&proc_root);
}

int mspproc_register_unregister_msp(struct i2c_client *c)
{
  int i;
  /*if client already exists it should be unregistered*/
  for(i=0;i<MAXMSPS;i++)
    {
      if (!private.proc_dir_entries[i]
	  ||private.proc_dir_entries[i]->data !=c) continue;
      dprk("client %s unregistered\n",c->name);
      delete_procfile(private.proc_dir_entries[i]);      
      private.proc_dir_entries[i]=0;
      private.nmsps--;
      if(private.nmsps<0)
	{
	  eprk("unregistered more than registered\n");
	  private.nmsps=0;
	  return -1;
	}
      return 0;
    }
  /*if client does not exist it should be registered*/
  for(i=0;i<MAXMSPS;i++)
    {
      if (private.proc_dir_entries[i]) continue;
      private.proc_dir_entries[i]=create_procfile(i,c);      
      if(!private.proc_dir_entries[i])
	{
	  eprk("could not register %s\n",c->name);
	  return -1;
	}
      private.nmsps++;
      dprk("wph: %s (%s) registered\n",
	   c->name,c->adapter->name); 
      return 0;
    }
  eprk("mspproc: too many msps\n");
  return -1;
}



#ifndef I2C_DRIVERID_MSPPROC
#define I2C_DRIVERID_MSPPROC I2C_DRIVERID_EXP3
#endif

static int mspproc_adapters(struct i2c_adapter *adap);
static int mspproc_clients(struct i2c_client *client);


static struct i2c_driver driver = {
	name: "msp3410d proc file driver",
        id: I2C_DRIVERID_MSPPROC /* I2C_DRIVERID_FIXME */,
	flags: I2C_DF_NOTIFY,
        attach_adapter: mspproc_adapters,
        detach_client: mspproc_clients,
};

static int mspproc_adapters(struct i2c_adapter *adap)
{
  int i;
 if (debug)
    printk("mspproc: adapter %s\n",adap->name);
  for (i=0; i<I2C_CLIENT_MAX; i++) {
    if (!adap->clients[i])
      continue;
    mspproc_clients(adap->clients[i]);
  }
  return 0;
}



static int mspproc_clients(struct i2c_client *client)
{
  /* TV card ??? */
  if (client->adapter->id != (I2C_ALGO_BIT | I2C_HW_B_BT848)) {
    dprk("mspproc: %s is not a tv card\n",
	     client->adapter->name);
    return -1;
  }
  dprk("mspproc: examining %s\n",client->name);
  
  if(client->driver->id!=I2C_DRIVERID_MSP3400)
    {
      dprk("mspproc: %s: driver->name is not an MSP i\n",
	     client->driver->name);
      return -1;
      
    }
  mspproc_register_unregister_msp(client);
  return 0;

}




int wph_init_module(void) 
{
  
  memset(&private,0,sizeof(private));
  private.nmsps=0;
  i2c_add_driver(&driver);
  return 0;
}

void wph_cleanup_module(void) 
{
  int i;
  printk("wph: cleanup called\n");
  i2c_del_driver(&driver);
  for(i=0;i<MAXMSPS;i++)
    {
      if(!private.proc_dir_entries[i]) continue;
      delete_procfile(private.proc_dir_entries[i]);      
      private.proc_dir_entries[i]=0;
      private.nmsps--;      
    }
  if(private.nmsps)
    eprk("private.nmsps=%d and should be 0\n",private.nmsps);
}


MODULE_AUTHOR("Wilfried Philips");
MODULE_DESCRIPTION("implements proc entry for msp3410D");
MODULE_PARM(debug,"i");

module_init(wph_init_module);
module_exit(wph_cleanup_module);

/*
 * Local variables:
 *  compile-command: "sudo make &&sudo make install && sudo depmod -a"
 * End:
 */
