/*
  
  File wphproc.c - /proc routines for msp34xx

  
  Copyright (c) 2000 W. Philips
  
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc.,   59 Temple Place, Suite 330 
  Boston, MA  02111-1307 USA 
*/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "wphprocraw.h"

struct msp_val_name_struct
{
    unsigned int val;
    char *name;
} msp_channel_source_modes[]={
{0x00, "FM/AM"},
{0x01, "NICAM"  }, 
{0x03, "none (MSPB/C: SBUS12)" }, 
{0x04, "none (MSPB/C: SBUS34)" }, 
{0x02, "SCART" }, 
{0x05, "I2S1"  }, 
{0x06, "I2S2"  },
{-1, ""}
};


struct msp_val_name_struct msp_channel_matrix_modes[]={
{0x00, "SOUNDA / LEFT / MSP-IF-CHANNEL2"},
{0x10, "RESET SOUNDB / RIGHT / MSP-IF-CHANNEL1"},
{0x20 , "STEREO"},
{0x30 , "MONO"},
{0x40 , "SUM / DIFF"},
{0x50 , "AB_XCHANGE"},
{0x60 , "PHASE_CHANGE_B"},
{0x70 , "PHASE_CHANGE_A"},
{0x80 , "A_ONLY"},
{0x90, "B_ONLY"},
{-1, ""}
 };


struct msp_dem_address_struct
{
  int addr;
  char *name;
  char t;
} msp_dem_addrs[]=
{
{0x0020, "DemShortProg", 'w'},
{0x0021, "AUTO_FM/AM", 'w'},
{0x00BB, "AD_CV", 'w'},
{0x0083, "MODE_REG", 'w'},
{0x0001, "FIR1", 'w'},
{0x0005, "FIR2", 'w'},
{0x0093, "DCO1_LO", 'w'},
{0x009B, "DCO1_HI", 'w'},
{0x00A3, "DCO2_LO", 'w' },
{0x00AB, "DCO2_HI", 'w'},
{0x001F, "PLL_CAPS", 'w'},
{0x007E, "ResultOfAutodetection", 'r'},
{0x0023, "C_AD_BITS", 'r'},
{0x0038, "ADD_BITS", 'r'},
{0x003E, "CIB_BITS", 'r'},
{0x0057, "ERROR_RATE", 'r'},
{0x0058, "CONC_CT", 'r'},
{0x0025, "FAWCT_IST", 'r'},
{0x021F, "PLL_CAPS", 'r'},
{0x021E, "AGC_GAIN", 'r'},
{0x0001, "Autodetection", 's' },
{0x0002, "M Dual-FM", 's'},
{0x0003, "B/G Dual-FM", 's'},
{0x0004, "D/K1 Dual-FM", 's'},
{0x0005, "D/K2 Dual-FM", 's'},
{0x0008, "B/G NICAM FM", 's'},
{0x0009, "L NICAM AM", 's'},
{0x000A, "I NICAM FM", 's'},
{0x000B, "D/K NICAM FM", 's'},
{-1, "" ,' '}
};


struct msp_dsp_address_struct
{
  int addr;
  char *name;
  char hl;
  char t;
  struct msp_val_name_struct *vn;
  unsigned short mask;
} msp_dsp_addrs[]=
{
{0x0000,	"Volume loudspeaker channel",  'H', 'w'},
{0x0000,	"Volume / Mode loudspeaker channel",  'L', 'w'},
{0x0001,	"Balance loudspeaker channel [L/R]",  'H', 'w'},
{0x0001,	" Balance Mode loudspeaker",  'L', 'w'},
{0x0002,	"Bass loudspeaker channel", 'H', 'w'},
{0x0003,	"Treble loudspeaker channel", 'H', 'w'},
{0x0004,	"Loudness loudspeaker channel",  'H', 'w'},
{0x0004,	"Loudness Filter Characteristic",  'L', 'w'},
{0x0005,	"Spatial effect strength loudspeaker ch",  'H', 'w'},
{0x0005,	"Spatial effect mode/customize",  'L', 'w'},
{0x0006,	"Volume headphone channel",  'H', 'w'},
{0x0006,	"Volume / Mode headphone channel",  'L', 'w'},
{0x0007,	"Volume / SCART1 channel",  'H', 'w'},
{0x0007,	"Volume / Mode SCART1 channel",  'L', 'w'},
{0x0008,	"Loudspeaker channel source",  'H', 'w',msp_channel_source_modes,0xff00},
{0x0008,	"Loudspeaker channel matrix",  'L', 'w',msp_channel_matrix_modes,0xff},
{0x0009,	"Headphone channel source",  'H', 'w',msp_channel_source_modes,0xff00},
{0x0009,	"Headphone channel matrix",  'L', 'w',msp_channel_matrix_modes,0xff},
{0x000A,	"SCART1 channel source",  'H', 'w',msp_channel_source_modes,0xff00},
{0x000A,	"SCART1 channel matrix",  'L', 'w',msp_channel_matrix_modes,0xff},
{0x000B,	"I2S channel source",  'H', 'w',msp_channel_source_modes,0xff00},
{0x000B,	"I2S channel matrix",  'L', 'w',msp_channel_matrix_modes,0xff},
{0x000C,	"Quasi-peak detector source",  'H', 'w',msp_channel_source_modes,0xff00},
{0x000C,	"Quasi-peak detector matrix",  'L', 'w',msp_channel_matrix_modes,0xff},
{0x000D,	"Prescale SCART",  'H', 'w'},
{0x000E,	"Prescale FM/AM",  'H', 'w'},
{0x000E,	"FM matrix",  'L', 'w'},
{0x000F,	"Deemphasis FM",  'H', 'w'},
{0x000F,	"Adaptive Deemphasis FM",  'L', 'w'},
{0x0010,	"Prescale NICAM",  'H', 'w'},
{0x0012,	"Prescale I2S2", 'H', 'w'},
{0x0013, "ACB Register (SCART Switching Facilities and Digital Control Output  Pins)",  0, 'w'},
{0x0014,      "Beeper", 0, 'w'},
{0x0015,      "Identification Mode", 'L', 'w'},
{0x0016,      "Prescale I2S1", 'H', 'w'},
{0x0017,      "FM DC Notch", 'L', 'w'},
{0x0020,      "Mode Tone Control", 'H', 'w'},
{0x0021,      "Equalizer loudspeaker ch. band 1", 'H', 'w'},
{0x0022,      "Equalizer loudspeaker ch. band 2", 'H', 'w'},
{0x0023,      "Equalizer loudspeaker ch. band 3", 'H', 'w'},
{0x0024,      "Equalizer loudspeaker ch. band 4", 'H', 'w'},
{0x0025,      "Equalizer loudspeaker ch. band 5", 'H', 'w'},
{0x0029,      "Automatic Volume Correction", 'H', 'w'},
{0x002C,      "Volume Subwoofer channel", 'H', 'w'},
{0x002D,      "Subwoofer Channel Corner Frequency", 'H', 'w'},
{0x002D,     "Subwoofer: Complementary High-pass", 'L', 'w'},
{0x0030, "Balance headphone channel [L/R]", 'H', 'w'},
{0x0030,	   "Balance Mode headphone",  'L', 'w'},
{0x0031,	   "Bass headphone channel", 'H', 'w'},
{0x0032,	   "Treble headphone channel", 'H', 'w'},
{0x0033,	   "Loudness headphone channel", 'H', 'w'},
{0x0033,	   "Loudness filter characteristic", 'L', 'w'},
{0x0040,	   "Volume SCART2 channel", 'H', 'w'},
{0x0040,	   "Volume / Mode SCART2 channel", 'L', 'w'},
{0x0041,	   "SCART2 channel source", 'H', 'w',msp_channel_source_modes,0xff},
{0x0041,	   "SCART2 channel matrix", 'L', 'w'},
{0x0018,	   "Stereo detection register", 'H', 'r'},
{0x0019,	   "Quasi-peak readout left", 0, 'r'},
{0x001A,	   "Quasi-peak readout right", 0, 'r'},
{0x001B,	   "DC level readout FM1/Ch2-L", 0, 'r'},
{0x001C,	   "DC level readout FM2/Ch1-R", 0, 'r'},
{0x001E,	   "MSP hardware version code", 'H', 'r'},
{0x001E,	   "MSP major revision code", 'L', 'r'},
{0x001F,	   "MSP product code", 'H', 'r'},
{0x001F,	   "MSP ROM version code", 'L', 'r'},
{-1, "" ,' '}
};



char *find_name_of_val(struct msp_val_name_struct*vn,int val,int mask)
{

  if(!vn) return "";
  switch(mask)
    {
    case 0xff: val &=0xff;
      break;
    case 0xff00: val >>= 8;
      break;
    default:
      break;
    }
  for(;vn->val!=-1;vn++)
    {
      if (vn->val==val) 
	return  vn->name;
    }
  return "";
}

int wph_read(int dev, int address)
{
  FILE *fp=fopen("/proc/" procname, "r+");
  datagram d;
  d.command=WPH_READ;
  d.dev=dev;
  d.address=address;
  if(!fp||(1!=fwrite(&d,sizeof(datagram),1,fp)))
    {
      fprintf(stderr,"error while writing\n");
      return 0;
    }
  fflush(fp);
  if(!fp||(1!=fread(&d,sizeof(datagram),1,fp)))
    {
      fprintf(stderr,"error while reading\n");
      return 0;
    }
  return d.value;
}


int wph_write(int dev, int address,unsigned int mask, unsigned int value)
{
  FILE *fp=fopen("/proc/" procname, "r+");
  datagram d;
  d.command=WPH_WRITE;
  d.dev=dev;
  d.address=address;
  d.mask=mask;
  d.value=value;
  if(!fp||(1!=fwrite(&d,sizeof(datagram),1,fp)))
    {
      fprintf(stderr,"error while writing\n");
      return 0;
    }
  fflush(fp);
  if(!fp||(1!=fread(&d,sizeof(datagram),1,fp)))
    {
      fprintf(stderr,"error while reading\n");
      return 0;
    }
  return d.value;
}




int printall(void)
{
  int len=0;  /* The number of bytes actually used */
  static int curlen=0;
  int clen;
  int val;
  char *name;
  static int called=0;
  int i,j,addr,hl;
  struct msp_val_name_struct*vn=0;
  unsigned int mask=0;
  char *valname=0;
  called++;
  for(i=0;msp_dsp_addrs[i].addr>=0;i++)
    {
      
      addr=msp_dsp_addrs[i].addr;
      name=msp_dsp_addrs[i].name;
      mask=msp_dsp_addrs[i].mask;
      vn=msp_dsp_addrs[i].vn;
      hl=msp_dsp_addrs[i].hl;
      val=wph_read(WPH_DSP,addr);
      valname=find_name_of_val(vn,val,mask);
      switch(hl)
	{
	case 'H':
	  printf("DSP%4.4xH=0x%-4.2x %s=%s\n",addr,val>>8,
			name,valname);
	  break;
	case 'L':
	  printf("DSP%4.4xL=0x%-4.2x %s=%s\n",addr,
			val&0xff,name,valname);
	  break;
	case 0:
	  printf("DSP%4.4x=0x%4.4x %s=%s\n",addr,val,
			name,valname);
	  break;
	}
	    }

}

/*
 tst get|set dsp|dem address [value mask]
 */
int main(int argc, char *argv[])
{
  int set=0;
  int get=0;
  int dsp=0;
  int dem=0;
  unsigned int value=0;
  unsigned int address=0;
  unsigned int mask=0;
  assert(argc>=4);
  if(!strcmp(argv[1],"set")) set=1;
  if(!strcmp(argv[1],"get")) get=1;
  assert((set&&argc==6)||(get&&argc==4));
  if(!strcmp(argv[2],"dsp")) dsp=1;
  if(!strcmp(argv[2],"dem")) dem=1;
  assert (dem||dsp);
  address=strtoul((argv[3]),0,0);
  if(set)
    {
      value=strtoul((argv[4]),0,0);
      mask=strtoul((argv[5]),0,0);
      printf("value=%4.4x mask=%4.4x\n",value,mask);
    }
  if(get)
    {
      value=wph_read((dsp)?WPH_DSP:WPH_DEM,address);
      printf("%s %4.4x=%4.4x\n",argv[2],address,value);
      return;
    }
  if(set)
    {
      value=wph_write((dsp)?WPH_DSP:WPH_DEM,address,mask,value);
      printf("%s %4.4x=%4.4x\n",argv[2],address,value);
      return;
    }
}


/*
 * Local variables:
 *  compile-command: "cc -g -o tst test.c"
 * End:
 */
