/*
  
  File wphcommand.c - 

  
  Copyright (c) 2000 W. Philips
  
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc.,   59 Temple Place, Suite 330 
  Boston, MA  02111-1307 USA 
*/
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include "kcompat24.h"
#include <linux/string.h>
#include <linux/proc_fs.h>
#include <linux/videodev.h>
#include <linux/i2c.h>
#include "wphprocraw.h"


#define I2C_MSP3400C       0x80
#define I2C_MSP3400C_DEM   0x10
#define I2C_MSP3400C_DFP   0x12

#define MAXMSPS 4

extern int debug=0;


void print_datagram(datagram *d)
{
  printk("datagram: %d %4.4x %4.4x %4.4x %4.4x %4.4x\n",
	 d->command,
	 d->dev,
	 d->address,
	 d->value,
	 d->mask,
	 d->status);
}

int  wph_mspset(struct i2c_client *client,int addr,int dev,int *val,int mask)
{
  int oldval=0;
  int t=0;
  int ret=0;
  switch(dev)
    {
    case WPH_DSP:
      t=I2C_MSP3400C_DFP;
      break;
    case WPH_DEM:
      t=I2C_MSP3400C_DEM;
      break;
    default:
      return -1;
    }
  if((mask&0xffff)!=0xffff)
    {
      oldval=msp3400c_read(client,t,addr);
    }
  printk("msp write: 0x%4.4x=0x%4.4x ->",addr,*val);
  *val=(oldval&~mask)|(*val&mask);
  printk("wph_mspset: 0x%4.4x\n",*val);
  ret=msp3400c_write(client,t ,addr, *val);
  return (ret== -1)?-1:0;
}



int  wph_mspget(struct i2c_client *client,int addr,int dev,int *val)
{

  int t=0;
  switch(dev)
    {
    case WPH_DSP:
      t=I2C_MSP3400C_DFP;
      break;
    case WPH_DEM:
      t=I2C_MSP3400C_DEM;
      break;
    default:
      return -1;
    }
  *val=msp3400c_read(client,t,addr);
  printk("msp read: 0x%4.4x=0x%4.4x\n",addr,*val);
  return (*val == -1)?-1:0;
}

 int wph_msp_command(struct i2c_client*client,datagram*d)
{
  printk("wph_msp_command\n");
  print_datagram(d);
    switch(d->command)
    {
    case WPH_READ:
      d->status=
	wph_mspget(client,d->address,
		   d->dev,&d->value);
      break;
    case WPH_WRITE:
      d->status=
	wph_mspset(client,d->address,d->dev,
		 &d->value,d->mask);
      break;
    default:
      printk("wph: illegal datagram command ignored\n");
      return sizeof(datagram);
    }
  return d->status;
}

/*protocol:
1. writing data to the device:
  set command=WPH_WRITE ; fill in address value and mask
  write datagram to /proc/mspraw
  read datagram from /proc/mspraw
  value contains the value of the address, status an error code

2. reading data from the msp
  set command=WPH_READ; fill in address 
  write datagram;
  read datagram
  and use value (mask is ignored)


 */


MODULE_AUTHOR("wph");
MODULE_DESCRIPTION("test");
MODULE_PARM(debug,"i");


EXPORT_SYMBOL(wph_msp_command);


int wph_init_module(void) 
{
  /*MOD_INC_USE_COUNT;*/
  printk("wph: init called\n");
	return 0;
}

void wph_cleanup_module(void) 
{
  printk("wph: cleanup called\n");
  /*  MOD_DEC_USE_COUNT;*/
	return ;
}

module_init(wph_init_module);
module_exit(wph_cleanup_module);

/*
 * Local variables:
 *  compile-command: "sudo make &&sudo make install && sudo depmod -a"
 * End:
 */
