/*
  
  File wphproc.c - /proc routines for msp34xx

  
  Copyright (c) 2000 W. Philips
  
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc.,   59 Temple Place, Suite 330 
  Boston, MA  02111-1307 USA 
*/
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/proc_fs.h>
#include <linux/videodev.h>
#include <linux/i2c.h>


#define I2C_MSP3400C       0x80
#define I2C_MSP3400C_DEM   0x10
#define I2C_MSP3400C_DFP   0x12

#define MAXMSPS 4

static struct wphprocdata
{
  int nmsps;
} private;

struct msp3400c * proc_msps[MAXMSPS];
struct i2c_client * proc_i2cs[MAXMSPS];

static int debug=0;



#define PREF " "

static  int snprintf(char*p,int size, char *fmt,...)
{
        va_list args;
        int len;
        char sbuffer[512];
        va_start(args, fmt);
        len = vsprintf(sbuffer, fmt, args);
        va_end(args);
	if(len>sizeof(sbuffer))
	  {
	    printk("wrote outside of buffer\n");
	  }
	if(size<=0)
	  return len;
	strncpy(p,sbuffer,(size<len)?size:len);
	if(len>size)
	  p[size-1]=0;
	if(len>size){
	  len=size;
	  sbuffer[len-1]='0';
	}
        return len;
}


int wphproc_register_msp(struct msp3400c *msp,struct i2c_client *c)
{
  if(private.nmsps<MAXMSPS)
    { 
      proc_i2cs[private.nmsps]=c;
      proc_msps[private.nmsps++]=msp;
      printk("registered proc%d 0x%x\n",private.nmsps,(unsigned int) msp);
      return 0;
    }
  else 
    {
      printk("cannot register proc 0x%x\n",(unsigned int)msp);
      return 0;
    }
}


int wphproc_unregister_msp(struct msp3400c *msp)
{
  int i;
  int j;
  for(i=0;i<private.nmsps;i++)
    {
      if (proc_msps[i]==msp) break;
    }
  j=i;
  if(i==private.nmsps)
    {
      printk("Trying to unregister non-registered msp\n");
      return -1;
    }
  for(;i<private.nmsps-1;i++)
    {
      proc_msps[i]=proc_msps[i+1];
    }

  printk("unregistered proc%d 0x%x\n",j,(unsigned int)msp);
  private.nmsps--;
  return 0;
}















struct msp_val_name_struct
{
    unsigned int val;
    char *name;
} msp_channel_source_modes[]={
{0x00, "FM/AM"},
{0x01, "NICAM"  }, 
{0x03, "none (MSPB/C: SBUS12)" }, 
{0x04, "none (MSPB/C: SBUS34)" }, 
{0x02, "SCART" }, 
{0x05, "I2S1"  }, 
{0x06, "I2S2"  },
{-1, ""}
};


struct msp_val_name_struct msp_channel_matrix_modes[]={
{0x00, "SOUNDA / LEFT / MSP-IF-CHANNEL2"},
{0x10, "RESET SOUNDB / RIGHT / MSP-IF-CHANNEL1"},
{0x20 , "STEREO"},
{0x30 , "MONO"},
{0x40 , "SUM / DIFF"},
{0x50 , "AB_XCHANGE"},
{0x60 , "PHASE_CHANGE_B"},
{0x70 , "PHASE_CHANGE_A"},
{0x80 , "A_ONLY"},
{0x90, "B_ONLY"},
{-1, ""}
 };


struct msp_dem_address_struct
{
  int addr;
  char *name;
  char t;
} msp_dem_addrs[]=
{
{0x0020, "DemShortProg", 'w'},
{0x0021, "AUTO_FM/AM", 'w'},
{0x00BB, "AD_CV", 'w'},
{0x0083, "MODE_REG", 'w'},
{0x0001, "FIR1", 'w'},
{0x0005, "FIR2", 'w'},
{0x0093, "DCO1_LO", 'w'},
{0x009B, "DCO1_HI", 'w'},
{0x00A3, "DCO2_LO", 'w' },
{0x00AB, "DCO2_HI", 'w'},
{0x001F, "PLL_CAPS", 'w'},
{0x007E, "ResultOfAutodetection", 'r'},
{0x0023, "C_AD_BITS", 'r'},
{0x0038, "ADD_BITS", 'r'},
{0x003E, "CIB_BITS", 'r'},
{0x0057, "ERROR_RATE", 'r'},
{0x0058, "CONC_CT", 'r'},
{0x0025, "FAWCT_IST", 'r'},
{0x021F, "PLL_CAPS", 'r'},
{0x021E, "AGC_GAIN", 'r'},
{0x0001, "Autodetection", 's' },
{0x0002, "M Dual-FM", 's'},
{0x0003, "B/G Dual-FM", 's'},
{0x0004, "D/K1 Dual-FM", 's'},
{0x0005, "D/K2 Dual-FM", 's'},
{0x0008, "B/G NICAM FM", 's'},
{0x0009, "L NICAM AM", 's'},
{0x000A, "I NICAM FM", 's'},
{0x000B, "D/K NICAM FM", 's'},
{-1, "" ,' '}
};


struct msp_dsp_address_struct
{
  int addr;
  char *name;
  char hl;
  char t;
  struct msp_val_name_struct *vn;
  unsigned short mask;
} msp_dsp_addrs[]=
{
{0x0000,	"Volume loudspeaker channel",  'H', 'w'},
{0x0000,	"Volume / Mode loudspeaker channel",  'L', 'w'},
{0x0001,	"Balance loudspeaker channel [L/R]",  'H', 'w'},
{0x0001,	" Balance Mode loudspeaker",  'L', 'w'},
{0x0002,	"Bass loudspeaker channel", 'H', 'w'},
{0x0003,	"Treble loudspeaker channel", 'H', 'w'},
{0x0004,	"Loudness loudspeaker channel",  'H', 'w'},
{0x0004,	"Loudness Filter Characteristic",  'L', 'w'},
{0x0005,	"Spatial effect strength loudspeaker ch",  'H', 'w'},
{0x0005,	"Spatial effect mode/customize",  'L', 'w'},
{0x0006,	"Volume headphone channel",  'H', 'w'},
{0x0006,	"Volume / Mode headphone channel",  'L', 'w'},
{0x0007,	"Volume / SCART1 channel",  'H', 'w'},
{0x0007,	"Volume / Mode SCART1 channel",  'L', 'w'},
{0x0008,	"Loudspeaker channel source",  'H', 'w',msp_channel_source_modes,0xff00},
{0x0008,	"Loudspeaker channel matrix",  'L', 'w',msp_channel_matrix_modes,0xff},
{0x0009,	"Headphone channel source",  'H', 'w',msp_channel_source_modes,0xff00},
{0x0009,	"Headphone channel matrix",  'L', 'w',msp_channel_matrix_modes,0xff},
{0x000A,	"SCART1 channel source",  'H', 'w',msp_channel_source_modes,0xff00},
{0x000A,	"SCART1 channel matrix",  'L', 'w',msp_channel_matrix_modes,0xff},
{0x000B,	"I2S channel source",  'H', 'w',msp_channel_source_modes,0xff00},
{0x000B,	"I2S channel matrix",  'L', 'w',msp_channel_matrix_modes,0xff},
{0x000C,	"Quasi-peak detector source",  'H', 'w',msp_channel_source_modes,0xff00},
{0x000C,	"Quasi-peak detector matrix",  'L', 'w',msp_channel_matrix_modes,0xff},
{0x000D,	"Prescale SCART",  'H', 'w'},
{0x000E,	"Prescale FM/AM",  'H', 'w'},
{0x000E,	"FM matrix",  'L', 'w'},
{0x000F,	"Deemphasis FM",  'H', 'w'},
{0x000F,	"Adaptive Deemphasis FM",  'L', 'w'},
{0x0010,	"Prescale NICAM",  'H', 'w'},
{0x0012,	"Prescale I2S2", 'H', 'w'},
{0x0013, "ACB Register (SCART Switching Facilities and Digital Control Output  Pins)",  0, 'w'},
{0x0014,      "Beeper", 0, 'w'},
{0x0015,      "Identification Mode", 'L', 'w'},
{0x0016,      "Prescale I2S1", 'H', 'w'},
{0x0017,      "FM DC Notch", 'L', 'w'},
{0x0020,      "Mode Tone Control", 'H', 'w'},
{0x0021,      "Equalizer loudspeaker ch. band 1", 'H', 'w'},
{0x0022,      "Equalizer loudspeaker ch. band 2", 'H', 'w'},
{0x0023,      "Equalizer loudspeaker ch. band 3", 'H', 'w'},
{0x0024,      "Equalizer loudspeaker ch. band 4", 'H', 'w'},
{0x0025,      "Equalizer loudspeaker ch. band 5", 'H', 'w'},
{0x0029,      "Automatic Volume Correction", 'H', 'w'},
{0x002C,      "Volume Subwoofer channel", 'H', 'w'},
{0x002D,      "Subwoofer Channel Corner Frequency", 'H', 'w'},
{0x002D,     "Subwoofer: Complementary High-pass", 'L', 'w'},
{0x0030, "Balance headphone channel [L/R]", 'H', 'w'},
{0x0030,	   "Balance Mode headphone",  'L', 'w'},
{0x0031,	   "Bass headphone channel", 'H', 'w'},
{0x0032,	   "Treble headphone channel", 'H', 'w'},
{0x0033,	   "Loudness headphone channel", 'H', 'w'},
{0x0033,	   "Loudness filter characteristic", 'L', 'w'},
{0x0040,	   "Volume SCART2 channel", 'H', 'w'},
{0x0040,	   "Volume / Mode SCART2 channel", 'L', 'w'},
{0x0041,	   "SCART2 channel source", 'H', 'w',msp_channel_source_modes,0xff},
{0x0041,	   "SCART2 channel matrix", 'L', 'w'},
{0x0018,	   "Stereo detection register", 'H', 'r'},
{0x0019,	   "Quasi-peak readout left", 0, 'r'},
{0x001A,	   "Quasi-peak readout right", 0, 'r'},
{0x001B,	   "DC level readout FM1/Ch2-L", 0, 'r'},
{0x001C,	   "DC level readout FM2/Ch1-R", 0, 'r'},
{0x001E,	   "MSP hardware version code", 'H', 'r'},
{0x001E,	   "MSP major revision code", 'L', 'r'},
{0x001F,	   "MSP product code", 'H', 'r'},
{0x001F,	   "MSP ROM version code", 'L', 'r'},
{-1, "" ,' '}
};



char *find_name_of_val(struct msp_val_name_struct*vn,int val,int mask)
{

  if(!vn) return "";
  switch(mask)
    {
    case 0xff: val &=0xff;
      break;
    case 0xff00: val >>= 8;
      break;
    default:
      break;
    }
  for(;vn->val!=-1;vn++)
    {
      if (vn->val==val) 
	return  vn->name;
    }
  return "";
}

void wph_msp3400c_interpret(int dev, int addr, int val, int rw)
{
  int i;
  static int called=0;
  int count=0;
  char *devstr;
  char *addrstr="";
  char *addrstr2="";
  char modestr=' ';
  char *valstr="";
  char *valstr2="";
  char *rwstring=rw?"write":"read";
  if(!called++)
    printk("WPHREG: START\n");
  switch (dev)
    {
    case I2C_MSP3400C_DEM:
      devstr="DEM";
      for(i=0;msp_dem_addrs[i].addr>=0;i++)
	{
	  if(msp_dem_addrs[i].addr==addr)
	    {
	      addrstr=msp_dem_addrs[i].name;
	      modestr=msp_dem_addrs[i].t;
	      break;
	    }
	}
      break;
    case I2C_MSP3400C_DFP:
      devstr="DSP";
      for(i=0;msp_dsp_addrs[i].addr>=0;i++)
	{
	  if(msp_dsp_addrs[i].addr==addr)
	    {
	      if(count)
		  {
		    addrstr2=msp_dsp_addrs[i].name;
		    if(msp_dsp_addrs[i].vn)
		      valstr2=find_name_of_val(msp_dsp_addrs[i].vn,val,
					       msp_dsp_addrs[i].mask);
		    break;
		  }
		else
		  {
		    count++;
		    addrstr=msp_dsp_addrs[i].name;
		    modestr=msp_dsp_addrs[i].t;
		    if(msp_dsp_addrs[i].vn)
		      valstr=find_name_of_val(msp_dsp_addrs[i].vn,val,
					      msp_dsp_addrs[i].mask);
		  }
	    }
	}
      break;
    default:
      devstr="";
    }
  printk("WPHREG: %d (%s) 0x%4.4x (%c:%s+%s) %s 0x%4.4x (%s+%s)\n",dev,devstr,addr,
	 modestr,addrstr,addrstr2,
	 rwstring,val,valstr,valstr2);
}













#if 0                                            
int procfile_read_OK(char *buffer, 
                  char **buffer_location, 
                  off_t offset, 
                  int buffer_length, 
                  int zero)
{
  int len=0;  /* The number of bytes actually used */
  int clen;
  static int called=0;
  int rem=buffer_length;
  int i;
  char *p=buffer;
  struct msp3400c*msp;
  *buffer_location=0;
  called++;
  if (offset > 0||!private.nmsps)
    return 0;
  
  for(i=0;i<private.nmsps;i++)
    {
      msp=proc_msps[i];
      clen=snprintf(p,rem,"count=%d %d 0x%x\n",called,msp->count,msp);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "simple=%d\n",msp->simple);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "nicam=%d\n",msp->nicam);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "mode=%d\n",msp->mode);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "wph_mode=%d: %s\n",msp->wph_mode,
		    msp->wph_mode_name);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "norm=%d\n",msp->norm);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "stereo=%d: ",msp->stereo);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=stereoprint(p,msp->stereo);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "watch_stereo=%d\n",msp->watch_stereo);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "nicam_on=%d\n",msp->nicam_on);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "main=%d\n",msp->main);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "second=%d\n",msp->second);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "left=%d\n",msp->left);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "right=%d\n",msp->right);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "bass=%d\n",msp->bass);
      p+=clen;rem-=clen;if(rem<=0) break;
      clen=snprintf(p,rem,PREF "treble=%d\n",msp->treble);
      p+=clen;rem-=clen;if(rem<=0) break;
    }
  /* Tell the function which called us where the 
   * buffer is */
  *buffer_location = 0;

  /* Return the length */
  len=p-buffer;
  if(len>buffer_length)len=buffer_length;
  return len;
}
#endif

static int msp3400c_reset(struct i2c_client *client)
{
        static char reset_off[3] = { 0x00, 0x80, 0x00 };
        static char reset_on[3]  = { 0x00, 0x00, 0x00 };

        i2c_master_send(client,reset_off,3);  /* XXX ignore errors here */
        if (3 != i2c_master_send(client,reset_on, 3)) {
		printk(KERN_ERR "msp3400: chip reset failed, penguin on i2c bus?\n");
                return -1;
	}
        return 0;
}

static int			
msp3400c_read(struct i2c_client *client, int dev, int addr)
{
	int err;

        unsigned char write[3];
        unsigned char read[2];
        struct i2c_msg msgs[2] = {
                { client->addr, 0,        3, write },
                { client->addr, I2C_M_RD, 2, read  }
        };
        write[0] = dev+1;
        write[1] = addr >> 8;
        write[2] = addr & 0xff;

	for (err = 0; err < 3;) {
		if (2 == i2c_transfer(client->adapter,msgs,2))
			break;
		err++;
		printk(KERN_WARNING "msp34xx: I/O error #%d (read 0x%02x/0x%02x)\n",
		       err, dev, addr);
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(HZ/10);
	}
	if (3 == err) {
		printk(KERN_WARNING "msp34xx: giving up, reseting chip. Sound will go off, sorry folks :-|\n");
		msp3400c_reset(client);
		return -1;
	}
	if(debug>2) wph_msp3400c_interpret(dev, addr, read[0] << 8 |read[1],0);
        return read[0] << 8 | read[1];
}


static int			/* wph: changed static to extern */
msp3400c_write(struct i2c_client *client, int dev, int addr, int val)
{
	int err;
        unsigned char buffer[5];

        buffer[0] = dev;
        buffer[1] = addr >> 8;
        buffer[2] = addr &  0xff;
        buffer[3] = val  >> 8;
        buffer[4] = val  &  0xff;

	for (err = 0; err < 3;) {
		if (5 == i2c_master_send(client, buffer, 5))
			break;
		err++;
		printk(KERN_WARNING "msp34xx: I/O error #%d (write 0x%02x/0x%02x)\n",
		       err, dev, addr);
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(HZ/10);
	}
	if (3 == err) {
		printk(KERN_WARNING "msp34xx: giving up, reseting chip. Sound will go off, sorry folks :-|\n");
		msp3400c_reset(client);
		return -1;
	}
	if(debug>2) wph_msp3400c_interpret(dev, addr, val,1);
	return 0;
}



char mymess[4096*4];


read_proc_t procfile_read;

int procfile_read(char *page, 
                  char **start, 
                  off_t off, 
                  int count, int *eof,
                  void *data)
{
  int len=0;  /* The number of bytes actually used */
  static int curlen=0;
  int clen;
  int val;
  char *name;
  static int called=0;
  int rem=sizeof(mymess);
  int i,j,addr,hl;
  char *p=mymess;
  struct i2c_client *client;
  struct msp_val_name_struct*vn=0;
  int mask=0;
  char *valname=0;
  *start=mymess;
  *eof=1;
  called++;
  if (!private.nmsps)
    return 0;
  if (off>0) 
    {
      *start=mymess+off;
      len=curlen-off;
      if(len<0) len=0;
      *eof=(len<count);
      return len;
    }
  for(j=0;j<private.nmsps;j++)
    {
      client=proc_i2cs[j];
      for(i=0;i<private.nmsps;i++)
	{
	  clen=snprintf(p,rem,"Demodulator:\n");
	  p+=clen;rem-=clen;if(rem<=0) goto end;
	  for(i=0;msp_dem_addrs[i].addr>=0;i++)
	    {
	      if(msp_dem_addrs[i].t!='r')
		continue;
	      addr=msp_dem_addrs[i].addr;
	      name=msp_dem_addrs[i].name;
	      val=0;
	      val=msp3400c_read(client,I2C_MSP3400C_DEM,addr);
	      valname=find_name_of_val(vn,val,mask);
	      clen=snprintf(p,rem,"DEM%4.4x=0x%4.4x %s=%s\n",addr,val,name,
			    valname);
	      p+=clen;rem-=clen;if(rem<=0) goto end;
	    }

	  clen=snprintf(p,rem,"DSP:\n");
	  p+=clen;rem-=clen;if(rem<=0) goto end;
	  for(i=0;msp_dsp_addrs[i].addr>=0;i++)
	    {
	      addr=msp_dsp_addrs[i].addr;
	      name=msp_dsp_addrs[i].name;
	      mask=msp_dsp_addrs[i].mask;
	      vn=msp_dsp_addrs[i].vn;
	      hl=msp_dsp_addrs[i].hl;
	      val=0;
	      val=msp3400c_read(client,I2C_MSP3400C_DFP,addr);
	      valname=find_name_of_val(vn,val,mask);
	      switch(hl)
		{
		case 'H':
		  clen=snprintf(p,rem,"DSP%4.4xH=0x%-4.2x %s=%s\n",addr,val>>8,
				name,valname);
		  break;
		case 'L':
		  clen=snprintf(p,rem,"DSP%4.4xL=0x%-4.2x %s=%s\n",addr,
				val&0xff,name,valname);
		  break;
		case 0:
		  clen=snprintf(p,rem,"DSP%4.4x=0x%4.4x %s=%s\n",addr,val,
				name,valname);
		  break;
		}
	      p+=clen;rem-=clen;if(rem<=0) goto end;
	    }
	  
	}
    }
 end:
  curlen=p-mymess;
  len=curlen;
  *eof=(len<count);
  return len;
}





#define procname "msp"


enum {ISDSP,ISDEM,ISILLEGAL} RTYPE;

enum {ISLOW,ISHIGH,ISWIDE} ADDRTYPE;


void  wph_mspset(int addr,int val,int rtype,int addrtype)
{
  int oldval=0;
  int t=0;
  struct i2c_client *client=proc_i2cs[0];
  switch(rtype)
    {
    case ISDSP:
      t=I2C_MSP3400C_DFP;
      break;
    case ISDEM:
      t=I2C_MSP3400C_DEM;
      break;
    default:
      return;
    }
  if(addrtype!=ISWIDE)
    {
      oldval=msp3400c_read(client,t,addr);
      val &=0xff;
    }

  printk("msp write: 0x%4.4x=0x%4.4x ->",addr,val);
  if(addrtype==ISLOW)
    val= (oldval & 0xff00)|(val);
  else if(addrtype==ISHIGH)
    val= (oldval & 0x00ff)|(val<<8);
  msp3400c_write(client,t ,addr, val);
  printk("0x%4.4x\n",val);
}

 int procfile_write(struct file *file, const char *buffer,
                           unsigned long count, void *data)
{
  const int str_size=256;
  char *str=kmalloc(str_size,GFP_KERNEL);
  char *hex=kmalloc(str_size,GFP_KERNEL);
  const char *dsp="DSP";
  const char *dem="DEM";
  char *found=0;
  char *found1=0;
  char *found2=0;
  int n=str_size-1;
  int type=0;
  int addr=0;
  int addrtype=0;
  int val=0;
  strncpy(hex,"0x",2);
  if(count<n) n=count;
  strncpy(str,buffer,n);
  str[n]=0;
  found1=strstr(str,dsp);
  found2=strstr(str,dem);
  if((found1-found2)<0)
    {type=ISDSP;found=found1+sizeof(dsp)-1;}
  else  if((found2-found1)<0)
    {type=ISDEM;found=found2+sizeof(dem)-1;}
  else
    type=ISILLEGAL;
  if(type!=ISILLEGAL)
    {
      if(found-str+5>=n)	/* need 4 digits and 1 of 'H' 'L' or '=' */
	type=ISILLEGAL;
    }
  else found=str+n;
  if(type!=ISILLEGAL)
    {
      strncpy(hex+2,found,4);	/* 4 hex digits */
      hex[6]=0;
      addr=simple_strtoul(hex,0,0);found+=4;
      addrtype=ISWIDE;
      if(type==ISDSP)
	{
	  if(found[0]=='H') addrtype=ISHIGH;
	  else if(found[0]=='L') addrtype=ISLOW;		  
	}
      found++;
      if(addrtype!=ISWIDE)found++; /* now found points after '=' */
      if (found-str>=n|| found[-1]!='=')
	type=ISILLEGAL;
    }
  if(type!=ISILLEGAL)
    {
      if((found-str+4>=n)
	 ||((found-str+6>=n)&&addrtype==ISWIDE)
	 ||found[0]!='0'
	 ||found[1]!='x')
	type=ISILLEGAL;
      else 
	{
	  int l=(addrtype==ISWIDE)?4:2;
	  found+=2;
	  strncpy(hex+2,found,l);	/*  4 or 2 hex digits */
	  hex[2+l]=0;
	  val=simple_strtoul(hex,0,0);found+=l;
	}
    }
  /* now skip to end of line, which is (dangerously) assumed to 
     fall within current str
   */
  { 
    char *f;
    f=strchr(found,'\n');
    if(!f) type=ISILLEGAL;
    else found=f+1;
  }
  if(found-str<count) count=found-str;
  if(count>=n)count=n;
  if(type!=ISILLEGAL)
    printk("wphproc: set: %s%4.4x%c=%4.4x\n",
	 (type==ISDEM)?dem:dsp,addr,
	 (addrtype==ISWIDE)?'W': ((addrtype==ISLOW)?'L':'H'),
	 val);
  wph_mspset(addr,val,type,addrtype);
  kfree(hex);
  kfree(str);
  return count;

}


struct proc_dir_entry Our_Proc_File = 
  {
    0, /* Inode number - ignore, it will be filled by 
        * proc_register[_dynamic]*/
    sizeof(procname)-1, /* Length of the file name*/
    procname, /* The file name*/
    S_IFREG | S_IRUGO | S_IWUGO, /* File mode - this is a regular 
				  * file which can be read and written 
				  * by its * owner, its group, and everybody
				  * else*/
    1,  /* Number of links (directories where the 
         * file is referenced)*/
    0, 0,  /* The uid and gid for the file - we give it 
            * to root*/
    80, /* The size of the file reported by ls.*/
    NULL, /* inode operations*/
    NULL, /*getinfo*/
    NULL, /*fill inode*/
    NULL,NULL,NULL, /*next, parent, subdir*/
    NULL, /*data*/
    procfile_read, /* The read function for this file, 
                    * the function called when somebody 
                    * tries to read something from it. */
    procfile_write,
    NULL /* We could have here a function to fill the 
          * file's inode, to enable us to play with 
          * permissions, ownership, etc. */
  }; 



/* Initialize the module - register the proc file */
int init_wphproc(void)
{
  /* Success if proc_register[_dynamic] is a success, 
   * failure otherwise. */
  /* In version 2.2, proc_register assign a dynamic 
   * inode number automatically if it is zero in the 
   * structure , so there's no more need for 
   * proc_register_dynamic
   */
  private.nmsps=0;
  printk("init_wph_proc\n");
#if 0
  Our_Proc_File.name="msp3400";
  Our_Proc_File.namelen=strlen(Our_Proc_File.name);
#endif
  return proc_register(&proc_root, &Our_Proc_File);
  /* proc_root is the root directory for the proc 
   * fs (/proc). This is where we want our file to be 
   * located. 
   */
  return 0;
}



/* Cleanup - unregister our file from /proc */
void cleanup_wphproc(void)
{
  proc_unregister(&proc_root, Our_Proc_File.low_ino);
}

MODULE_AUTHOR("wph");
MODULE_DESCRIPTION("test");
MODULE_PARM(debug,"i");


EXPORT_SYMBOL(cleanup_wphproc);
EXPORT_SYMBOL(init_wphproc);
EXPORT_SYMBOL(wphproc_register_msp);
EXPORT_SYMBOL(wphproc_unregister_msp);
EXPORT_SYMBOL(wph_msp3400c_interpret);

int wph_init_module(void) 
{
  MOD_INC_USE_COUNT;
  init_wphproc();
	return 0;
}

void wph_cleanup_module(void) 
{
  cleanup_wphproc();
  MOD_DEC_USE_COUNT;
	return ;
}

module_init(wph_init_module);
module_exit(wph_cleanup_module);

/*
 * Local variables:
 *  compile-command: "sudo make &&sudo make install && sudo depmod -a"
 * End:
 */
